package org.qtunes.auth.control;

import org.qtunes.core.*;
import org.qtunes.auth.*;
import org.qtunes.db.*;
import org.qtunes.db.Field;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/**
 * The ControlMethodInvoker takes a method description to its "invoke" method
 * and invokes the matching {link ControlMethod}.
 */
public class ControlMethodInvoker {

    private final ServiceContext context;
    private final Auth auth;

    /**
     * Create a new ControlMethodInvoker
     * @param qtunes the qtunes;
     */
    public ControlMethodInvoker(ServiceContext context) {
        this.context = context;
        this.auth = context.getService(Auth.class);
    }

    /**
     * Given the name of an Object, return the Object itself.
     * By default this maps "qtunes" and "db" to the main
     * server and the Database, and maps Service by its
     * service-name
     * @param objectname the name of the Object
     * @return the Object, or null if not found
     */
    public Object getObject(String objectname) {
        if (objectname.equalsIgnoreCase("qtunes")) {
            return context;
        } else {
            return context.getService(Service.class, "name=\""+objectname+"\"");
        }
    }

    /**
     * Invoke a {@link ControlMethod} on an Object. The object will be searched
     * for a matching method, and the arguments - specified in the "args" map as either
     * positional values like <code>{1:"value1", 2:"value2"}</code>, or by name eg
     * <code>{"track":"1", "index":0}</code>. This class will attempt to cast the
     * parameters (say converting "Album" into {@link Field#Album} an arrays into
     * {@link BitSet} objects), and if that succeeds then it will call the method
     * and return the return value or {@link Void#TYPE} if the method returns
     * no value.
     *
     * @param qtunes the QTunes
     * @param objectname the name of the Object, to be passed in to {@link #getObject}
     * @param methodname the name of the method on the Object to invoke
     * @param args the parameters. These will be matched against the paramaters of the ControlMethod
     * by type, using the {@link #convert} method to type match if necessary
     * @return the Object returned from the method, or {@link Void#TYPE} if the method has no return
     */
    public Object invoke(ServiceContext context, String objectname, List<Call> calls) throws Throwable {
        Object object = getObject(objectname);
        if (object==null) {
            throw new IllegalArgumentException("Unknown object \""+objectname+"\"");
        }
        String nametodate = objectname;
        for (int k=0;k<calls.size();k++) {
            String methodname = calls.get(k).getMethod();
            Map args = calls.get(k).getParameters();
            Method[] methods = object.getClass().getMethods();
            Throwable storede = null;
            boolean found = false;
            for (int i=0;i<methods.length;i++) {
                Method method = methods[i];
                ControlMethod annot;
                if (method.getName().equals(methodname) && (annot=getControlMethodAnnotation(method))!=null) {
                    String authrequired = annot.auth();
                    String[] paramnames = annot.params();
                    Class[] paramclasses = method.getParameterTypes();
                    Object[] invokeargs = new Object[paramclasses.length];
                    try {
                        // Special case - single Map argument called "properties" gets the whole map
                        if (paramnames.length==1 && paramnames[0].equals("properties") && Map.class.isAssignableFrom(paramclasses[0])) {
                            invokeargs[0] = args;
                        } else {
                            for (int j=0;j<invokeargs.length;j++) {
                                String paramname = j < paramnames.length ? paramnames[j] : "param-"+j;
                                Object value = args.get(paramname);
                                if (value==null && !args.containsKey(paramname)) {
                                    value = args.get(Integer.valueOf(j));
                                    if (value==null) {
                                        throw new IllegalArgumentException("Parameter \""+paramname+"\" not specified for "+objectname+"."+methodname);
                                    }
                                }
                                try {
                                    invokeargs[j] = convert(value, paramclasses[j]);
                                } catch (Exception e) {
                                    String classname = paramclasses[j].getName();
                                    classname = classname.substring(classname.lastIndexOf(".") + 1);
                                    String valstring = value instanceof String ? "\""+value+"\"" : value.toString();
                                    throw new IllegalArgumentException("Can't convert "+valstring+" into "+classname+" for parameter "+j+" \""+paramname+"\"", e);
                                }
                            }
                        }
                        if (method != null) {
                            if (Session.getCurrentSession().isAllowed(authrequired)) {
                                object = method.invoke(object, invokeargs);
                                if (method.getReturnType()==Void.TYPE) {
                                    if (k==calls.size()-1) {
                                        return Void.TYPE;
                                    } else {
                                        throw new IllegalStateException("Method \""+nametodate+"."+methodname+"\" returns null");
                                    }
                                } else {
                                    found = true;
                                    nametodate = nametodate + "." + calls.get(k);
                                }
                            } else {
                                throw new IllegalStateException("Method \""+nametodate+"."+methodname+"\" requires \""+authrequired+"\" permission");
                            }
                        }
                    } catch (InvocationTargetException e) {
                        return e.getTargetException();
                    } catch (Exception e) {
                        storede = e;
                    }
                }
            }
            if (storede!=null) {
                throw storede;
            } else if (!found) {
                throw new IllegalArgumentException("Unknown method \""+methodname+"\" with "+args.size()+" parameters on \""+nametodate+"\"");
            }
        }
        return object;
    }

    /**
     * Get the ControlMethod annotation on the supplied method, possibly
     * from an interface or superclass definition
     */
    @SuppressWarnings("unchecked")
    private static ControlMethod getControlMethodAnnotation(Method m) {
        String name = m.getName();
        Class<?>[] types = m.getParameterTypes();
        Class cl = m.getDeclaringClass();
        ControlMethod annot = null;
        do {
            try {
                m = cl.getMethod(name, types);
                annot = (ControlMethod)m.getAnnotation(ControlMethod.class);
            } catch (Exception e) {}
            Class[] interfaces = cl.getInterfaces();
            for (int i=0;annot==null && i<interfaces.length;i++) {
                try {
                    m = interfaces[i].getMethod(name, types);
                    annot = (ControlMethod)m.getAnnotation(ControlMethod.class);
                } catch (Exception e) {}
            }
            cl = cl.getSuperclass();
        } while (annot==null && cl!=null);
        return annot;
    }

    protected Object convert(Object o, Class paramclass) throws Exception {
        if (o==null || isAssignable(paramclass, o.getClass())) {
            return o;
        } else if (paramclass.isArray()) {
            Object array;
            if (o instanceof List) {
                List newo = (List)o;
                array = Array.newInstance(paramclass.getComponentType(), newo.size());
                for (int i=0;i<newo.size();i++) {
                    Array.set(array, i, convert(newo.get(i), paramclass.getComponentType()));
                }
            } else {
                int len = Array.getLength(o);
                array = Array.newInstance(paramclass.getComponentType(), len);
                for (int i=0;i<len;i++) {
                    Array.set(array, i, convert(Array.get(o, i), paramclass.getComponentType()));
                }
            }
            return array;
        } else if (paramclass==Integer.TYPE || paramclass==Integer.class) {
            if (o instanceof Number) {
                return new Integer(((Number)o).intValue());
            } else {
                return Integer.decode(o.toString());
            }
        } else if (paramclass==Float.TYPE || paramclass==Float.class) {
            if (o instanceof Float) {
                return new Float(((Number)o).floatValue());
            } else {
                return Float.valueOf(o.toString());
            }
        } else if (paramclass==Boolean.class || paramclass==Boolean.TYPE) {
            if ("true".equalsIgnoreCase(o.toString()) || "1".equals(o.toString())) {
                return Boolean.TRUE;
            } else if ("false".equalsIgnoreCase(o.toString()) || "0".equals(o.toString())) {
                return Boolean.FALSE;
            }
        } else if (paramclass==BitSet.class) {
            int[] values = (int[])convert(o, Array.newInstance(Integer.TYPE, 0).getClass());
            BitSet bitset = new BitSet();
            for (int i=0;i<values.length;i++) {
                bitset.set(values[i]);
            }
            return bitset;
        } else if (paramclass==String.class) {
            return o.toString();
        } else if (paramclass==FieldMap.class) {
            FieldMap map = new FieldMap();
            for (Iterator i = ((Map)o).entrySet().iterator();i.hasNext();) {
                Map.Entry e = (Map.Entry)i.next();
                AbstractField<?> field = Field.getByName((String)e.getKey());
                if (field!=null && field instanceof Field) {
                    map.putUntyped((Field)field, e.getValue());
                }
            }
            return map;
        } else if (paramclass==Session.class) {
            return auth.getSession(((Integer)o).intValue());
        } else if (paramclass==File.class) {
            return new File(o.toString());
        } else if (AbstractField.class.isAssignableFrom(paramclass)) {
            return Field.getByName(o.toString());
        } else if (Playlist.class.isAssignableFrom(paramclass)) {
            Database db = context.getService(Database.class);
            if (o.equals("")) {
                return null;
            } else if (o instanceof String) {
                Playlist playlist = db.getPlaylist((String)o);
                if (playlist!=null) {
                    return playlist;
                } else {
                    return db.createSmartPlaylist((String)o);
                }
            } else if (o instanceof BitSet) {
                ManualPlaylist playlist = db.createManualPlaylist();
                BitSet bs = (BitSet)o;
                for (int i=bs.nextSetBit(0);i>=0;i=bs.nextSetBit(i)) {
                    Track track = db.getTrack(i);
                    if (track!=null) {
                        playlist.add(track);
                    }
                }
                return playlist;
            } else {
                int[] values = (int[])convert(o, Array.newInstance(Integer.TYPE, 0).getClass());
                ManualPlaylist playlist = db.createManualPlaylist();
                for (int i=0;i<values.length;i++) {
                    Track track = db.getTrack(values[i]);
                    if (track!=null) {
                        playlist.add(track);
                    }
                }
                return playlist;
            }
        } else if (paramclass==Class.class) {
            return Class.forName(o.toString());
        } else if (paramclass==Service.class) {
            return context.getService(Service.class, "name=\""+o+"\"");
        }
        throw new IllegalArgumentException();
    }

    @SuppressWarnings("unchecked")
    private static boolean isAssignable(Class c1, Class c2) {
        if (c1.isAssignableFrom(c2)) {
            return true;
        } else if (c1==Integer.TYPE) {
            return c2==Integer.class || c2==Short.class || c2==Short.TYPE || c2==Byte.class || c2==Byte.TYPE || c2==Character.class || c2==Character.TYPE;
        } else if (c1==Float.TYPE) {
            return c2==Float.class || c2==Integer.class || c2==Integer.TYPE || c2==Short.class || c2==Short.TYPE || c2==Byte.class || c2==Byte.TYPE || c2==Character.class || c2==Character.TYPE;
        } else if (c1==Double.TYPE) {
            return c2==Long.TYPE || c2==Long.class || c2==Double.class || c2==Float.TYPE || c2==Float.class || c2==Integer.class || c2==Integer.TYPE || c2==Short.class || c2==Short.TYPE || c2==Byte.class || c2==Byte.TYPE || c2==Character.class || c2==Character.TYPE;
        } else {
            return false;
        }
    }

    public List<Method> getControlMethods(Object o) {
        List<Method> methods = new ArrayList<Method>();
        Class cl = o.getClass();
        do {
            Method[] m = cl.getMethods();
            for (int j=0;j<m.length;j++) {
                if (m[j].isAnnotationPresent(ControlMethod.class)) {
                    methods.add(m[j]);
                }
            }
            Class[] interfaces = cl.getInterfaces();
            for (int i=0;i<interfaces.length;i++) {
                m = interfaces[i].getMethods();
                for (int j=0;j<m.length;j++) {
                    if (m[j].isAnnotationPresent(ControlMethod.class)) {
                        methods.add(m[j]);
                    }
                }
            }
            cl = cl.getSuperclass();
        } while (cl!=null);
        return methods;
    }

}
