package org.qtunes.auth.spi;

import org.qtunes.core.*;
import org.qtunes.auth.*;
import java.util.*;
import java.net.*;

/**
 * The Auth class handles creation of Sessions and determining what rights
 * those Sessions have. Each QTunes object has only one Auth service
 * Different Services may register an {@link Authorizer}
 * with this object: when a Session is created by a call to {@link #login},
 * each Authorizer for the matching Service is queried to see what "allowances"
 * it gives it.
 */
public class AuthImpl implements Auth {

    private Set<Authorizer> authorizers;
    private Map<Integer,Session> sessions;
    private ServiceContext context;
    private Thread expirer;
    private int defaultvalidity = 300;
    private Random random;

    public AuthImpl() {
        random = new Random();
        sessions = new HashMap<Integer,Session>();
        authorizers = new HashSet<Authorizer>();
        expirer = new Thread("qTunes-auth-SessionExpirer") {
            public void run() {
                while (true) {
                    if (!sessions.isEmpty()) {
                        List<Session> copy = new ArrayList<Session>(sessions.size());
                        synchronized(sessions) {
                            copy.addAll(sessions.values());
                        }
                        long now = System.currentTimeMillis() / 1000;
                        for (int i=0;i<copy.size();i++) {
                            Session session = copy.get(i);
                            if (session.getExpiryTime() < now) {
                                logout(session);
                            }
                        }
                    }
                    try {
                        Thread.sleep(5000);
                    } catch(InterruptedException e) { }
                }
            }
        };
        expirer.setDaemon(true);
        expirer.start();
    }

    public void startService(ServiceContext context) {
        this.context = context;
        String[] keys = context.getPropertyNames();
        for (int i=0;i<keys.length;i++) {
            if (keys[i].startsWith("auth.")) {
                String value = context.getProperty(keys[i]);
                int ix = value.indexOf(" ");
                String[] allowed = value.substring(0, ix).split(",");
                String filter = value.substring(ix).trim();
                addAuthorizer(new BasicAuthorizer(new SessionMatcher(filter), allowed));
            }
        }
        try {
            defaultvalidity = Integer.valueOf(context.getProperty("ttl"));
        } catch (Exception e) {}
    }

    public void stopService(ServiceContext context) {
    }

    public ServiceContext getContext() {
        return context;
    }

    public Session getSession(int id) {
        synchronized(sessions) {
            return sessions.get(Integer.valueOf(id));
        }
    }

    @ControlMethod (
        params = {"properties"},
        help = "Create a new session"
    )
    public Session login(Map properties) {
        Set allowed = new HashSet();
        Map<String,Object> props = new HashMap<String,Object>();
        for (Iterator i = properties.entrySet().iterator();i.hasNext();) {
            Map.Entry e = (Map.Entry)i.next();
            if (e.getKey() instanceof String) {
                props.put((String)e.getKey(), e.getValue());
            }
        }

        Service service = null;
        if (props.get("service") instanceof Service) {
            service = (Service)props.get("service");
        } else {
            throw new IllegalArgumentException("Missing \"service\" property");
        }

        synchronized(sessions) {
            int id;
            do {
                id = (random.nextInt()) & 0x7FFFFFFF;
            } while (sessions.containsKey(Integer.valueOf(id)));

            Session session = new SessionImpl(this, id, service, props, defaultvalidity);
            if (session.isAllowed("ttl") && props.containsKey("ttl")) {
                try {
                    int ttl = Integer.valueOf(props.remove("ttl").toString());
                    session = new SessionImpl(this, id, service, props, ttl);
                } catch (Exception e) { }
            }
            if (session.isAllowed("*")) {
                sessions.put(new Integer(id), session);
                context.info("Creating session "+session);
            } else {
                session = Session.NONE;
            }
            return session;
        }
    }

    public void logout(Session session) {
        if (session!=null && session!=Session.NONE) {
            synchronized(sessions) {
                if (sessions.remove(Integer.valueOf(session.getId()))!=null) {
                    session.expire();
                    context.fireEvent("expire", new Object[] { "session", session });
                    context.info("Expiring session "+session);
                }
            }
        }
    }

    public Collection<String> getAllowances(Session session) {
        Set<String> allowances = new HashSet<String>();
        synchronized(authorizers) {
            for (Iterator<Authorizer> i = authorizers.iterator();i.hasNext();) {
                Authorizer auth = i.next();
                Collection<String> match = auth.getAllowances(session);
                if (match != null) {
                    allowances.addAll(match);
                }
            }
        }
        return allowances;
    }

    public void addAuthorizer(Authorizer auth) {
        synchronized(authorizers) {
            authorizers.add(auth);
        }
    }

    public void removeAuthorizer(Authorizer auth) {
        synchronized(authorizers) {
            authorizers.remove(auth);
        }
    }

    public void logoutSessions(Service service) {
        List<Session> copy = new ArrayList<Session>(sessions.size());
        synchronized(sessions) {
            copy.addAll(sessions.values());
        }
        for (int i=0;i<copy.size();i++) {
            if (copy.get(i).getService()==service) {
                logout(copy.get(i));
            }
        }
    }

    public Map<String,Object> reportState() {
        return null;
    }

}
