package org.qtunes.core.util;

import java.io.*;
import java.util.*;
import java.text.Normalizer;

public class FileUtils {

    private static InheritableThreadLocal<File> root = new InheritableThreadLocal<File>() {
        @Override protected File initialValue() {
            try {
                return new File(".").getCanonicalFile();
            } catch (IOException e) {
                throw new IllegalStateException(e);
            }
        };
    };

    private FileUtils() {
    }

    /**
     * Set the File root for this thread and its children
     */
    public static void setRoot(File file) throws IOException {
        if (!file.getCanonicalFile().isDirectory()) {
            throw new IOException("\""+file+"\" is not a directory");
        }
        root.set(file.getCanonicalFile());
    }

    /**
     * Get the File root for this thread and its children
     */
    public static File getRoot() {
        return root.get();
    }

    /**
     * Get the absolute path of the specified File relative to the current root
     */
    public static File getAbsoluteFile(File file) throws IOException {
        if (file.isAbsolute()) {
            return file.getCanonicalFile();
        } else {
            return (new File(root.get(), file.getPath())).getCanonicalFile();
        }
    }

    /**
     * Given a root and a file, return a new File representing the canonical
     * relative path from root to file
     */
    public static File getRelativeFile(File file) throws IOException {
        List<String> rootbits = new ArrayList<String>(Arrays.asList(split(root.get().getCanonicalPath(), File.separatorChar)));
        List<String> filebits = new ArrayList<String>(Arrays.asList(split(file.getCanonicalPath(), File.separatorChar)));
        int i = 0;
        while(i < rootbits.size() && i < filebits.size() && rootbits.get(i).equals(filebits.get(i))) {
            i++;
        }
        StringBuilder sb = new StringBuilder();
        int j = i;
        while (i < rootbits.size()) {
            sb.append("..");
            sb.append(File.separatorChar);
            i++;
        }
        while (j < filebits.size()) {
            sb.append(filebits.get(j++));
            if (j < filebits.size()) {
                sb.append(File.separatorChar);
            }
        }
        if (sb.length() == 0) {
            sb.append('.');
        }
        return new File(sb.toString());
    }

    public static boolean equals(File file1, File file2) throws IOException {
        File absfile1 = getAbsoluteFile(file1);
        File absfile2 = getAbsoluteFile(file2);
        String p1 = absfile1.getPath();
        String p2 = absfile2.getPath();
        if (System.getProperty("os.name").equals("Mac OS X")) {
            p1 = Normalizer.normalize(p1, Normalizer.Form.NFC);
            p2 = Normalizer.normalize(p2, Normalizer.Form.NFC);
        }
        return p1.equals(p2);
    }

    /**
     * Like String.split but not regex sensitive, for splitting D:\\path\\
     * type input
     */
    private static String[] split(String in, char sep) {
        int count = 1;
        for (int i=0;i<in.length();i++) {
            if (in.charAt(i) == sep) {
                count++;
            }
        }
        String[] out = new String[count];
        count = 0;
        int j = 0;
        for (int i=0;i<in.length();i++) {
            if (in.charAt(i) == sep) {
                out[j++] = in.substring(count, i);
                count = i+1;
            }
        }
        out[j] = in.substring(count);
        return out;
    }

}
