package org.qtunes.core.util;

import java.lang.reflect.Array;
import java.util.*;
import java.io.*;

/**
 * Class to read/write objects as JSON.
 * Can be overridden to support custom serialization. For writing this should involve
 * overriding the {@link #getCustomWrite} method like so:
 * <pre class="code">
 * public Object getCustomWrite(Object o) {
 *     if (o instanceof Widget) {
 *         Map m = new HashMap();
 *         m.put("type", "widget");
 *         m.put("name", widget.getName());
 *         return m;
 *     } else {
 *         return super.getCustomWrite(o);
 *     }
 * }
 * </pre>
 * and for reading the {@link #getCustomRead} methods:
 * <pre class="code">
 * public Object getCustomRead(Map m) {
 *     if ("widget".equals(m.get("type"))) {
 *         return Widget.getByName((String)m.get("name"));
 *     } else {
 *         return super.getCustomRead();
 *     }
 * }
 * </pre>
 */
public class JSONSerializer {

    private boolean laxkey;

    public void setLaxKeyQuoting(boolean lax) {
        this.laxkey = lax;
    }

    /**
     * Serialize the object and return the serialized version as a String
     * @param object the Object to serialize
     */
    public String write(Object object) {
        try {
            StringBuffer sb = new StringBuffer();
            write(object, sb);
            return sb.toString();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Serialize the object and write the serialized version to the specified {@link Appendable}
     * @param o the Object to serialize
     * @param sb the Appendable to write to
     */
    public void write(Object o, Appendable sb) throws IOException {
        write(o, sb, false, false);
    }

    /**
     * As for {@link #write(Object,Appendable) but compound objects are synchronized on before iterating
     * @param o the Object to serialize
     * @param sb the Appendable to write to
     */
    public void writeSynchronized(Object o, Appendable sb)  throws IOException {
        write(o, sb, false, true);
    }

    private void write(Object o, Appendable sb, boolean iskey, boolean synchronize) throws IOException {
        if (o instanceof StringBuffer || o instanceof StringBuilder) {
            o = o.toString();
        }
        if (o == null) {
            sb.append("null");
        } else if (o instanceof JSONSerializable {
            ((JSONSerializable)o).write(sb, synchronize);
        } else if (o instanceof String) {
            String string = (String)o;
            if (string.length() == 0) {
                sb.append("\"\"");
            } else {
                char c = 0;
                int len = string.length();
                Appendable origsb = sb;
                boolean escape = false;
                if (iskey && laxkey) {
                    sb = new StringBuilder();
                } else {
                    sb.append('"');
                }
                for (int i = 0; i < len; i++) {
                    char b = c;
                    c = string.charAt(i);
                    switch (c) {
                    case '\\':
                    case '"':
                        sb.append('\\');
                        sb.append(c);
                        escape = true;
                        break;
                    case '/':
                        if (b == '<') {
                            sb.append('\\');
                        }
                        sb.append(c);
                        escape = true;
                        break;
                    case '\b':
                        sb.append("\\b");
                        escape = true;
                        break;
                    case '\t':
                        sb.append("\\t");
                        escape = true;
                        break;
                    case '\n':
                        sb.append("\\n");
                        escape = true;
                        break;
                    case '\f':
                        sb.append("\\f");
                        escape = true;
                        break;
                    case '\r':
                        sb.append("\\r");
                        escape = true;
                        break;
                    default:
                        if (c < ' ' || (c >= '\u0080' && c < '\u00a0') || (c >= '\u2000' && c < '\u2100')) {
                            String t = "000" + Integer.toHexString(c);
                            sb.append("\\u" + t.substring(t.length() - 4));
                            escape = true;
                        } else {
                            sb.append(c);
                            if (!((c>='a' && c<='z') || (c>='A' && c<='Z') || (i>0 && ((c>='0' && c<='9') || c=='.' || c=='-')) || c=='_')) {
                                escape = true;
                            }
                        }
                    }
                }
                if (origsb != sb) {
                    if (escape) {
                        origsb.append('"');
                        origsb.append((CharSequence)sb);
                        origsb.append('"');
                    } else {
                        origsb.append((CharSequence)sb);
                    }
                    sb = origsb;
                } else {
                    sb.append('"');
                }
            }
        } else if (o instanceof Boolean) {
            sb.append(o.toString());
        } else if (o instanceof Number) {
            Number n = (Number)o;
            if (n instanceof Double && (((Double)n).isInfinite() || ((Double)n).isNaN())) {
                throw new IllegalArgumentException("Infinite or NaN");
            } else if (o instanceof Float && (((Float)o).isInfinite() || ((Float)o).isNaN())) {
                throw new IllegalArgumentException("Infinite or NaN");
            } else {
                String s = n.toString();
                if (s.indexOf('.') > 0 && s.indexOf('e') < 0 && s.indexOf('E') < 0) {
                    while (s.endsWith("0")) {
                        s = s.substring(0, s.length() - 1);
                    }
                    if (s.endsWith(".")) {
                        s = s.substring(0, s.length() - 1);
                    }
                }
                sb.append(s);
            }
        } else if (o.getClass().isArray()) {
            sb.append("[");
            if (synchronize) {
                synchronized(o) {
                    int len = Array.getLength(o);
                    for (int i=0;i<len;i++) {
                        if (i!=0) {
                            sb.append(",");
                        }
                        write(Array.get(o, i), sb, false, synchronize);
                    }
                }
            } else {
                int len = Array.getLength(o);
                for (int i=0;i<len;i++) {
                    if (i!=0) {
                        sb.append(",");
                    }
                    write(Array.get(o, i), sb, false, synchronize);
                }
            }
            sb.append("]");
        } else if (o instanceof Map) {
            sb.append("{");
            if (synchronize) {
                synchronized(o) {
                    boolean first = true;
                    for (Iterator i = ((Map)o).entrySet().iterator();i.hasNext();) {
                        Map.Entry e = (Map.Entry)i.next();
                        if (e.getKey() instanceof String && e.getValue()!=null) {
                            if (first) {
                                first = false;
                            } else {
                                sb.append(",");
                            }
                            write(e.getKey(), sb, true, synchronize);
                            sb.append(":");
                            write(e.getValue(), sb, false, synchronize);
                        }
                    }
                }
            } else {
                boolean first = true;
                for (Iterator i = ((Map)o).entrySet().iterator();i.hasNext();) {
                    Map.Entry e = (Map.Entry)i.next();
                    if (e.getKey() instanceof String && e.getValue()!=null) {
                        if (first) {
                            first = false;
                        } else {
                            sb.append(",");
                        }
                        write(e.getKey(), sb, true, synchronize);
                        sb.append(":");
                        write(e.getValue(), sb, false, synchronize);
                    }
                }
            }
            sb.append("}");
        } else if (o instanceof Collection) {
            sb.append("[");
            Collection list = (Collection)o;
            boolean first = true;
            if (synchronize) {
                synchronized(list) {
                    for (Iterator i=list.iterator();i.hasNext();) {
                        if (!first) {
                            sb.append(",");
                        }
                        first = false;
                        write(i.next(), sb, false, synchronize);
                    }
                }
            } else {
                for (Iterator i=list.iterator();i.hasNext();) {
                    if (!first) {
                        sb.append(",");
                    }
                    first = false;
                    write(i.next(), sb, false, synchronize);
                }
            }
            sb.append("]");
        } else {
            write(getCustomWrite(o), sb, false, synchronize);
        }
    }

    /**
     * Get an object that can be serialized to represent the specified Object.
     * This method is called when there is no default way to serialize
     * the parameter - by default it throws an Exception saying the object
     * cannot be serialized. It can be overridden by subclasses.
     * @param object an unserializable Object
     * @return an Object which can be used in it's place
     */
    public Object getCustomWrite(Object object) throws IllegalArgumentException {
        throw new IllegalArgumentException("Unable to serialize "+object);
    }

    public Object readAsProbableString(String s) throws IOException {
        try {
            return read(new StringReader(s));
        } catch (IOException e) {
            try {
                return read(new StringReader("\""+s+"\""));
            } catch (IOException e2) {
                throw e;
            }
        }
    }

    /**
     * Parse a JSON serialized String and return the Object it represents
     */
    public Object read(String s) throws IOException {
        return read(new StringReader(s));
    }

    private Object ENDARRAY = "']'";
    private Object ENDMAP = "'}'";
    private Object COMMA = "','";
    private Object COLON = "':'";
    private Object NULL = "null";

    /**
     * Parse a JSON serialized object from the Reader and return the Object it represents
     */
    public Object read(Reader reader) throws IOException {
        if (!reader.markSupported()) {
            reader = new BufferedReader(reader);
        }
        Object o = readToken(reader, null, false);
        int c;
        if ((c=stripBlanks(reader))!=-1) {
            throw new IOException("Unexpected trailing characters: c="+c);
        }
        return o;
    }

    private Object readToken(Reader reader, Object additional, boolean iskey) throws IOException {
        Object out = null;
        int c = stripBlanks(reader);
        if (c < 0) {
            throw new EOFException();
        } else if (c=='[') {
            List<Object> list = new ArrayList<Object>();
            Object o;
            while (true) {
                o = readToken(reader, list.isEmpty() ? ENDARRAY : null, false);
                if (o==ENDARRAY) {
                    break;
                } else {
                    list.add(o==NULL ? null : o);
                }
                if ((o = readEnd(reader))==ENDARRAY) {
                    break;
                } else if (o!=COMMA) {
                    throw new IOException("Unexpected token "+o);
                }
            }
            out = getCustomRead(list);
        } else if (c=='{') {
            Map<String,Object> map = new LinkedHashMap<String,Object>();
            Object o;
            while (true) {
                Object key = readToken(reader, map.isEmpty() ? ENDMAP : null, true);
                if (key==ENDMAP) {
                    break;
                }
                if (!(key instanceof String)) {
                    throw new IOException("Invalid object key "+key);
                }
                if ((o = readEnd(reader))!=COLON) {
                    throw new IOException("Unexpected token "+o);
                }
                map.put((String)key, readToken(reader, null, false));
                if ((o = readEnd(reader))==ENDMAP) {
                    break;
                } else if (o!=COMMA) {
                    throw new IOException("Unexpected token "+o);
                }
            }
            out = getCustomRead(map);
        } else if ((c>='0' && c<='9') || c=='-') {
            StringBuffer sb = new StringBuffer();
            boolean flt = false;
            reader.mark(1);
            sb.append((char)c);
            while ((c=reader.read()) >='0' && c<='9') {
                reader.mark(1);
                sb.append((char)c);
            }
            if (c=='.') {
                flt = true;
                sb.append((char)c);
                reader.mark(1);
                while ((c=reader.read()) >='0' && c<='9') {
                    reader.mark(1);
                    sb.append((char)c);
                }
            }
            if (c=='e' || c=='E') {
                flt = true;
                sb.append((char)c);
                reader.mark(1);
                if ((c=reader.read())=='+' || c=='-' || (c>='0' && c<='9')) {
                    reader.mark(1);
                    sb.append((char)c);
                    while ((c=reader.read()) >='0' && c<='9') {
                        reader.mark(1);
                        sb.append((char)c);
                    }
                }
            }
            if (flt) {
                try {
                    out = Double.valueOf(sb.toString());
                } catch (NumberFormatException e) {
                    throw new IOException("Invalid number", e);
                }
            } else {
                try {
                    out = Integer.valueOf(sb.toString());
                } catch (NumberFormatException e) {
                    try {
                        out = Long.valueOf(sb.toString());
                    } catch (NumberFormatException e2) {
                        throw new IOException("Invalid number", e2);
                    }
                }
            }
            reader.reset();
        } else if (c=='"') {
            StringBuffer sb = new StringBuffer();
            boolean inquotes = true;
            while ((c=reader.read()) >= 0) {
                if (c=='"') {
                   inquotes = false;
                   break;
                } else if (c=='\\') {
                    c = reader.read();
                    if (c=='\\' || c=='\"' || c=='/') {
                        sb.append((char)c);
                    } else if (c=='b') {
                        sb.append('\b');
                    } else if (c=='f') {
                        sb.append('\f');
                    } else if (c=='n') {
                        sb.append('\n');
                    } else if (c=='r') {
                        sb.append('\r');
                    } else if (c=='t') {
                        sb.append('\t');
                    } else if (c=='u') {
                        int c1 = reader.read();
                        int c2 = reader.read();
                        int c3 = reader.read();
                        int c4 = reader.read();
                        if (c1>='0' && c1<='9' && c2>='0' && c2<='9' && c3>='0' && c3<='9' && c4>='0' && c4<='9') {
                            sb.append(Integer.decode("0x"+((char)c1)+((char)c2)+((char)c3)+((char)c4)));
                        } else {
                            throw new IOException("Invalid unicode escape "+c1+" "+c2+" "+c3+" "+c4);
                        }
                    }
                } else if (!Character.isISOControl(c)) {
                    sb.append((char)c);
                } else {
                    throw new IOException("Invalid string character 0x"+c);
                }
            }
            if (inquotes) {
                throw new IOException("Unterminated string \""+sb);
            } else {
                out = sb.toString();
            }
        } else if ((c=='t' || c=='f' || c=='n') && !iskey) {
            StringBuffer sb = new StringBuffer();
            sb.append((char)c);
            reader.mark(1);
            while ((c=reader.read())>=0 && c>='a' && c<='z' && sb.length() < 5) {
                sb.append((char)c);
                reader.mark(1);
            }
            reader.reset();
            if (sb.toString().equals("true")) {
                out = Boolean.TRUE;
            } else if (sb.toString().equals("false")) {
                out = Boolean.FALSE;
            } else if (sb.toString().equals("null")) {
                out = null;
            } else {
                throw new IOException("Invalid token \""+sb+"\"");
            }
        } else if (c==']' && additional==ENDARRAY) {
            out = ENDARRAY;
        } else if (c=='}' && additional==ENDMAP) {
            out = ENDMAP;
        } else if (iskey && laxkey) {
            StringBuffer sb = new StringBuffer();
            sb.append((char)c);
            reader.mark(1);
            while ((c=reader.read()) >= 0 && ((c>='a' && c<='z') || (c>='A' && c<='Z') || (c>='0' && c<='9') || c=='_' || c=='-' || c=='.')) {
                
                reader.mark(1);
                sb.append((char)c);
            }
            reader.reset();
            out = sb.toString();
        } else {
            out = readCustomToken(c, reader);
            if (out==null) {
                throw new IOException("Unexpected token "+((char)c));
            }
        }
//        System.out.println("-> read(1) "+out);
        return out;
    }

    private Object readEnd(Reader reader) throws IOException, IllegalArgumentException {
        Object out;
        int c = stripBlanks(reader);
        if (c==']') {
            out = ENDARRAY;
        } else if (c=='}') {
            out = ENDMAP;
        } else if (c==',') {
            out = COMMA;
        } else if (c==':') {
            out = COLON;
        } else if (c==-1) {
            throw new EOFException();
        } else {
            throw new IOException("Unexpected character '"+((char)c)+"'");
        }
//        System.out.println("--> read(2) "+out);
        return out;
    }

    private static int stripBlanks(Reader reader) throws IOException {
        int c;
        while ((c=reader.read()) >= 0 && (c==' ' || c=='\n' || c=='\t' || c=='\r')) {
//            System.out.println("-> skip(1)="+((char)c));
        }
        if (c=='/') {
            reader.mark(1);
            if ((c=reader.read())=='*') {
                StringBuilder comment = new StringBuilder();
                comment.append("/*");
                int lastc = reader.read();
                while ((c=reader.read())>=0 && c!='/' && lastc!='*') {
                    comment.append((char)lastc);
                    lastc = c;
                }
                if (lastc < 0) throw new EOFException("Unterminated comment");
                comment.append("*/");
//                System.out.println("Read Comment \""+comment+"\"");
                return stripBlanks(reader);
            } else {
                reader.reset();
            }
        }
        return c;
    }

    /**
     * Subclasses may override this method to return a more specific
     * type of object than the Map specified. Roughly the opposite
     * of {@link #getCustomWrite}.
     * @param map a generic Map
     * @return the Object that the Map represents.
     */
    protected Object getCustomRead(Map map) {
        return map;
    }

    /**
     * Subclasses may override this method to return a more specific
     * type of object than the List specified. Roughly the opposite
     * of {@link #getCustomWrite}.
     * @param list a generic List
     * @return the Object that the Map represents.
     */
    protected Object getCustomRead(List list) {
        return list;
    }

    protected Object readCustomToken(int c, Reader reader) {
        return null;
    }

    /**
     * An object can implement this interface if it wants to control
     * how it's written to the JSON stream
     */
    public static final interface JSONSerializable {
        public void write(Appendable a, boolean synchronize) throws IOException
    }

    /*
    public static void main(String[] args) throws Exception {
        JSONSerializer json = new JSONSerializer();
        Object o = json.read(args[0]);
        String s = json.write(o);
        System.out.println(args[0]);
        System.out.println(o);
        System.out.println(s);
    }
    */
}
