package org.qtunes.daap;

import java.io.*;

/**
 * A Block is what the DAAP/DACP protocol is constructed of. May be a List, String or Number
 */
abstract class Block {

    private int code;

    /**
     * Set the numeric Block code, a 32-bit integer
     */
    void setCode(int code) {
        this.code = code;
    }

    int getCode() {
        return code;
    }

    /**
     * Get the name of the Block - a String version of the code
     */
    String getName() {
        return Blocks.codeToName(code);
    }

    /**
     * Get the length of the Block content when written to an OutputStream
     */
    abstract int length();

    /**
     * Write the content of the Block to the specified OutputStream
     */
    abstract void writeContent(OutputStream out) throws IOException;

    /**
     * Read the Block content from the specified InputStream
     */
    abstract void read(InputStream in, int length) throws IOException;

    /**
     * Write the Block to the specified OutputStream
     */
    final void write(OutputStream out) throws IOException {
        int code = getCode();
        out.write(code>>24);
        out.write(code>>16);
        out.write(code>>8);
        out.write(code>>0);
        int len = length();
        out.write(len>>24);
        out.write(len>>16);
        out.write(len>>8);
        out.write(len>>0);
        writeContent(out);
    }

    abstract void dump(Appendable sb, String prefix) throws IOException;

    public final void dump(Appendable out) throws IOException {
        dump(out, "\n");
    }

    public final String dump() {
        final int MAX = 1000;
        Appendable a = new Appendable() {
            StringBuffer sb = new StringBuffer();
            int len;

            public String toString() {
                return len <= sb.length() ? sb.toString() : sb.toString()+"...("+len+" chars)";
            }

            public Appendable append(char c) {
                len++;
                if (sb.length() < MAX) {
                    sb.append(c);
                }
                return this;
            }

            public Appendable append(CharSequence csq, int start, int end) {
                return append(csq.subSequence(start, end));
            }

            public Appendable append(CharSequence csq) {
                len += csq.length();
                if (sb.length() < MAX) {
                    if (sb.length() + csq.length() > MAX) {
                        csq = csq.subSequence(0, MAX - sb.length());
                    }
                    sb.append(csq);
                }
                return this;
            }
        };
        try {
            dump(a, null);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
        return a.toString();
    }
}
