package org.qtunes.daap;

import org.qtunes.core.*;
import org.qtunes.db.*;
import org.qtunes.auth.*;
import org.qtunes.web.*;
import org.qtunes.player.*;
import org.qtunes.zeroconf.*;
import java.net.*;
import java.io.*;
import java.util.*;
import java.util.concurrent.*;
import java.security.*;

/**
 * This Service handles DAAP and DACP requests via a WebServer, and manages
 * the announcement of this service via ZeroConf. It's used for streaming
 * music to iTunes, or to allow a Player to be remotely controlled by the
 * Remote app on an Apple iPhone. Only one DaapServer is required to service
 * multiple simultaneous clients - one is required for sharing, and when
 * playing the correct Player to control can be identified by the Remote
 * that's in use.
 */
public class DaapServer implements Service {
   
    static final int APRO_VERSION=0x00030000; // iTunes 9.0=3/8, firefly=3/0
    static final int MPRO_VERSION=0x00020000; // iTunes 9.0=2/6, firefly=2/0
    static final int AESV_VERSION=0x00030001; // iTunes 9.0=2/6, firefly=2/0

    private ServiceContext context;
    private WebService webservice;
    private RemoteManager remotemanager;
    private ServiceRegistrar serviceregistrar;
    private Map<Player,PlayerState> playerstates;
    private Auth auth;
    private Database database;
    private RootHandler roothandler;
    private String password;
    private volatile boolean sharing;
    private long uniqueid;
    private Map<Playlist,List<Group>> grouplists;
    private List<Group> allgroups = new ArrayList<Group>();
    private Listener statelistener;
    private Revision updaterevision = new Revision(this, 2);          // Arbitrary >1

    public DaapServer() {
        new DaapField();       // Init class
    }

    /**
     * Return true if this DaapServer is sharing it's collection
     * to iTunes, as set by {@link #setSharing}
     */
    public boolean isSharing() {
        return sharing;
    }

    /**
     * Set whether this DaapServer is sharing it's collection to iTunes
     */
    public void setSharing(boolean sharing) {
        this.sharing = sharing;
        roothandler.setServeDaap(sharing);
        serviceregistrar.setSharing(sharing);
    }

    long getUniqueID() {
        return uniqueid;
    }

    /**
     * Get the Player
     * IMPORTANT: requires a Session with a valid login, or returns null
     */
    Player getPlayer() {
        Session session = Session.getCurrentSession();
        String guid = (String)session.getProperties().get("guid");
        Player player = remotemanager.getPlayer(guid);
        if (player == null) {
            // Seen this happen once, can't reproduce so far
            throw new IllegalStateException("No player for "+guid);
        }
        return player;
    }

    /**
     * Return the Revision being requested by "/ctrl-int/playstatusupdate"
     * IMPORTANT: requires a Session with a valid login, or returns null
     */
    Revision getPlayRevision() {
        Session session = Session.getCurrentSession();
        String guid = (String)session.getProperties().get("guid");
        Player player = remotemanager.getPlayer(guid);
        PlayerState state = playerstates.get(player);
        return state.getRevision();
    }

    /**
     * Return the Now Playing playlist
     * IMPORTANT: requires a Session with a valid login, or returns null
     */
    private DaapPlaylist getNowPlayingPlaylist() {
        Session session = Session.getCurrentSession();
        String guid = (String)session.getProperties().get("guid");
        Player player = remotemanager.getPlayer(guid);
        if (player == null) {
            return null;
        } else {
            PlayerState state = playerstates.get(player);
            return state.getPlaylist();
        }
    }

    /**
     * Return the Revision being requested by "/update"
     */
    Revision getUpdateRevision() {
        return updaterevision;
    }

    /**
     * Get the List of DaapPlaylists that are in use with this
     * Server
     */
    synchronized List<DaapPlaylist> getPlaylists() {
        List<DaapPlaylist> out = new ArrayList<DaapPlaylist>();
        out.add(DaapPlaylist.getBasePlaylist(this));
        out.add(DaapPlaylist.getMusicPlaylist(this));
        for (String name : getDatabase().getPlaylists()) {
            Playlist playlist = getDatabase().getPlaylist(name);
            DaapPlaylist apl = DaapPlaylist.getInstance(this, name, playlist);
            out.add(apl);
        }
        return out;
    }

    /**
     * Get the DaapPlaylist with the specified index
     */
    DaapPlaylist getPlaylist(long index) {
        List<DaapPlaylist> playlists = getPlaylists();
        playlists.add(getNowPlayingPlaylist());
        for (int i=0;i<playlists.size();i++) {
            DaapPlaylist pl = playlists.get(i);
            if (pl.getIndex()==index || pl.getPersistentId()==index) {
                return pl;
            }
        }
        return null;
    }

    /**
     * Get the DaapPlaylist with the specified name
     */
    DaapPlaylist getPlaylist(String name) {
        List<DaapPlaylist> playlists = getPlaylists();
        playlists.add(getNowPlayingPlaylist());
        for (int i=0;i<playlists.size();i++) {
            if (name.equals(playlists.get(i).getName())) {
                return playlists.get(i);
            }
        }
        return null;
    }

    synchronized List<Group> getGroupList(Playlist filter) {
        if (grouplists==null) {
            grouplists = new HashMap<Playlist,List<Group>>();
        }
        List<Group> grouplist = grouplists.get(filter);
        if (grouplist==null) {
            grouplist = Group.createGroupList(this, filter);
            grouplists.put(filter, grouplist);
        }
        return grouplist;
    }

    void addGroup(Group group) {
        synchronized(allgroups) {
            allgroups.add(group);
            group.setIndex(allgroups.size());
        }
    }

    Group getGroup(int index) {
        synchronized(allgroups) {
            return allgroups.get(index-1);
        }
    }

    String getPassword() {
        return password;
    }

    Database getDatabase() {
        return database;
    }

    Auth getAuth() {
        return auth;
    }

    ServiceRegistrar getServiceRegistrar() {
        return serviceregistrar;
    }

    RemoteManager getRemoteManager() {
        return remotemanager;
    }

    ServiceContext getServiceContext() {
        return context;
    }

    WebService getWebService() {
        return webservice;
    }

    String getServerName() {
        return context.getServerName();
    }

    //-----------------------------------------------------------------------------

    public void startService(ServiceContext context) {
        this.context = context;
        database = context.getService(Database.class);
        auth = context.getService(Auth.class);
        playerstates = new ConcurrentHashMap<Player,PlayerState>();
        uniqueid = getUniqueID(this);

        password = context.getProperty("password");
        String port = context.getProperty("port");
        webservice = context.getService(WebService.class, port==null ? null : "port='"+port+"'");
        roothandler = new RootHandler(this);
        webservice.setHandler("/", roothandler);

        context.addListener(statelistener = new Listener() {
            public void handleEvent(String type, Map properties) {
                Service service = (Service)properties.get("service");
                if (service instanceof Player) {
                    Player player = (Player)service;
                    if (type.equals("started")) {
                        PlayerState state = new PlayerState(player, DaapServer.this);
                        playerstates.put(player, state);
                        state.start();
                        roothandler.setServeDacp(true);
                    } else if (type.equals("stopped")) {
                        PlayerState state = playerstates.remove(player);
                        state.stop();
                        roothandler.setServeDacp(!playerstates.isEmpty());
                    } else {
                        PlayerState state = playerstates.get(player);
                        if (state != null) {    // Might get event before we've started
                            state.bumpPlayer();
                        }
                    }
                } else if (service instanceof Database) {
                    for (Iterator<PlayerState> i = playerstates.values().iterator();i.hasNext();) {
                        i.next().bumpDatabase();
                    }
                }
            }
        });
        context.addListener(updaterevision);

        remotemanager = new RemoteManager(context);
        serviceregistrar = new ServiceRegistrar(this, webservice.getAddress());
        remotemanager.setServiceRegistrar(serviceregistrar);
        serviceregistrar.setRemoteManager(remotemanager);
        remotemanager.start();
        serviceregistrar.start();

        setSharing(!"false".equals(context.getProperty("sharing")));
    }

    public void stopService(ServiceContext context) {
        auth.logoutSessions(this);
        remotemanager.cancel();
        serviceregistrar.cancel();
        updaterevision.cancel();
        for (Iterator<PlayerState> i = playerstates.values().iterator();i.hasNext();) {
            i.next().stop();
            i.remove();
        }
        DaapPlaylist.free(this);
        context.removeListener(statelistener);
        context.removeListener(updaterevision);
        webservice.setHandler("/", null);
    }

    public ServiceContext getContext() {
        return context;
    }

    public Map<String,Object> reportState() {
        Map<String,Object> map = new LinkedHashMap<String,Object>();
        map.put("sharing", sharing);
        map.put("password", password!=null);
        map.put("pendingremotes", remotemanager.getPendingRemotes());
        return map;
    }

    static long getUniqueID(Service service) {
        String value = service.getContext().getServerName()+"\n"+service.getContext().getServiceName();
        try {
            MessageDigest digest = MessageDigest.getInstance("MD5");
            digest.update(value.getBytes("UTF-8"));
            byte[] b = digest.digest();
            long out = 0;
            for (int i=0;i<b.length;i++) {
                out ^= (long)b[i]<<((i&7)*8);
            }
            out &= 0x3FFFFFFFFFFFFFFFl;
            return out;
        } catch (NoSuchAlgorithmException e) {
            throw new RuntimeException(e);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Add a new Remote pairing
     * @param name the name of the Remote
     * @param player the Player it's being paired with
     * @param pin the 4-digit pin entered by the user to match that on the Remote
     */
    public void addRemote(String name, Player player, String pin) {
        remotemanager.addRemote(name, player, pin);
    }

    /**
     * Unpair a previously paired Remote
     * @param name the name of the Remote
     * @param player the Player it was paired with
     */
    public void removeRemote(String name, Player player) {
        remotemanager.removeRemote(name, player);
    }

}
