package org.qtunes.db;

import org.qtunes.core.*;
import java.io.*;
import java.util.*;

/**
 * The Database manages the Tracks and Playlists. How it
 * works is down to the implementation but in general the
 * rules are it must handle locking, as it will be read/
 * written to by multiple threads.
 */
public interface Database extends Service {

    /**
     * Create a new Track entry. The track will initially be created in the temporary
     * store - whether this method runs an "update" or an "insert" depends on the
     * {@link Matcher}, which will try to match the fields in the {@link FieldMap}
     * against the existing tracks. This may be done by Filename, by Inode or by 
     * checksum of the audio data (for example). The fields are then stored
     * and the Track returned.
     */
    public Track create(FieldMap map);

    /**
     * Return a read-only Collection with all the Tracks in the Database
     * in arbirary order.
     */
    @ControlMethod (
        params = {},
        help = "Return a Collection with all the Tracks in the Database"
    )
    public Collection<Track> getTracks();

    /**
     * Return a List of the tracks in the database with the specified sort
     * order and filter. This list will be regenerated on each call so should
     * be called sparingly. The returned list may be modified
     */
    public List<Track> getTracks(String sort, String filter);

    /**
     * Return the revision number of the Database
     */
    @ControlMethod (
        params = {},
        help = "Return the revision number of the Database"
    )
    public int getRevision();

    /**
     * Return the number of tracks in this Database
     */
    @ControlMethod (
        params = {},
        help = "Return the number of tracks in this Database"
    )
    public int size();

    /**
     * Get the track with the specified track index
     */
    public Track getTrack(int tracknumber);

    /**
     * Commit any pending database changes (new or modified tracks) to the database.
     */
    @ControlMethod (
        params = {},
        auth = "edit",
        help = "Commit any pending database changes (new or modified tracks) to the database."
    )
    public void commit();

    /**
     * Get the Playlist with the specified name. The special name "*"
     * means the Playlist that matches all tracks in the database
     */
    @ControlMethod (
        params = {"name"},
        auth = "*",
        help = "Get the Playlist with the specified name"
    )
    public Playlist getPlaylist(String name);

    /**
     * Get the list of Playlist names
     */
    @ControlMethod (
        params = {},
        auth = "*",
        help = "Get the list of Playlist names"
    )
    public Collection<String> getPlaylists();

    /**
     * Store (possibly overwriting) the playlist with the specified name.\
     * If the value is null the playlist will be removed.
     */
    @ControlMethod (
        params = {"name","playlist"},
        auth = "playlist",
        help = "Store (possibly overwriting) the playlist with the specified name"
    )
    public void putPlaylist(String name, Playlist playlist);

    /**
     * Update the fields on the specified Track. The track file will be
     * updated immediately, the database only on the next commit
     */
    @ControlMethod (
        params = {"track","changes"},
        auth = "edit",
        help = "Update the fields on the specified Track. The track file will be updated immediately, the database only on the next commit"
    )
    public void updateTrack(int tracknumber, FieldMap changes) throws IOException;

    /**
     * Create a new ManualPlaylist
     */
    public ManualPlaylist createManualPlaylist();

    /**
     * Create a new ManualPlaylist
     */
    public SmartPlaylist createSmartPlaylist(String filter);

    /** 
     * Lock or unlock the database for edits. Use sparingly!
     */
    public void lock(boolean hold);

}
