package org.qtunes.db;

import org.qtunes.core.*;
import java.util.*;
import java.io.*;
import java.nio.*;
import javax.sound.sampled.*;

/**
 A Track is a single Track. It's created by the {@link Database#create} method, which in turn gets
 * its values from a {@link org.qtunes.ff.TrackReader}. A Track is made up of a number of {@link AbstractField}
 * classes, and it can be converted to audio sample by calling {@link #getAudioInputStream}. Some
 * types of {@link org.qtunes.ff.TrackReader} are also instances of {@link org.qtunes.ff.TrackWriter}, and if that's the case
 * the Track can be updated by calling {@link Database#updateTrack}.
 */
public interface Track {

    /**
     * Get the index of this track in the database
     */
    public int getIndex();

    /**
     * Get the value of the specified Field. The return type
     * depends on the Field but should be a String, Integer, Boolean,
     * Float, Long, int[] or byte[], or it may be null. The other
     * get methods can be called to type the return value correctly.
     */
    @ControlMethod (
        params = {"field"},
        help = "Return the named Field from this Track"
    )
    public <T> T get(AbstractField<T> field);

    /**
     * Get the File this Track was loaded from
     */
    public File getFile();

    /**
     * Get the {@link AudioInputStream} for this track. Don't
     * forget to close the stream once you're done.
     */
    public AudioInputStream getAudioInputStream() throws IOException;

    /**
     * Get a {@link ByteBuffer} containing the image. This maps
     * the file into memory, so the returned ByteBuffer should
     * be copied if the return value is not to be discarded
     * immediately.
     */
    public ByteBuffer getImageBuffer() throws IOException;

    /**
     * Get an Iterator that moves over all the {@link Field} objects
     * in this Track
     */
    public Iterator<Field<?>> getFields();

    /**
     * Delete the track from the database
     */
    public void delete();

    /**
     * Update the fields on this Track. The track file will be
     * updated immediately, the database only on the next commit
     */
    @ControlMethod (
        params = {"changes"},
        auth = "edit",
        help = "Update the fields on this Track"
    )
    public void update(FieldMap changes) throws IOException;

}
