package org.qtunes.db.spi.simple;

import org.qtunes.core.*;
import org.qtunes.db.*;
import java.util.*;
import java.io.*;

/**
 * A Simple Playlist that has a manually managed list of tracks
 */
public class ManualPlaylistImpl implements ManualPlaylist {

    private final DatabaseImpl db;
    private String name;
    private final List<Track> tracks;
    private int revision, lastdbrevision, uniqueid;

    ManualPlaylistImpl(DatabaseImpl db, int uniqueid, List<Track> tracks) {
        this.db = db;
        this.uniqueid = uniqueid;
        this.tracks = tracks==null ? new ArrayList<Track>() : tracks;
    }

    public void setName(String name) {
        this.name = name;
    }

    public String getName() {
        return name;
    }

    public int getUniqueID() {
        return uniqueid;
    }

    public int getRevision() {
        return revision;
    }

    private void touch() {
        revision++;
        if (name!=null) {
            db.putPlaylist(name, this);         // Save on any change
        }
    }

    /**
     * Add the tracks specified in this Playlist
     * @param playlist the tracks to add
     */
    public synchronized void addAll(Playlist playlist) {
        tracks.addAll(playlist.getTracks());
        touch();
    }

    /**
     * Add the specified tracks to this Playlist, but only if they're not duplicates.
     * @param playlist the tracks to add
     */
    public synchronized void mergeAll(Playlist playlist) {
        List<Track> newtracks = playlist.getTracks();
        newtracks.removeAll(tracks);
        if (tracks.addAll(newtracks)) {
            touch();
        }
    }

    /**
     * Add the specified tracks to this Playlist at the specified point
     * @param playlist the tracks to add
     * @param index the insertion position in the list
     */
    public synchronized void addAll(Playlist playlist, int index) {
        if (index==tracks.size()) {
            addAll(playlist);
        } else {
            tracks.addAll(index, playlist.getTracks());
            touch();
        }
    }

    /**
     * Add the specified tracks to this Playlist at the specified point,
     * but only if they're not duplicates
     * @param playlist the tracks to add
     * @param index the insertion position in the list
     */
    public synchronized void mergeAll(Playlist playlist, int index) {
        if (index==tracks.size()) {
            mergeAll(playlist);
        } else {
            List<Track> newtracks = playlist.getTracks();
            newtracks.removeAll(tracks);
            if (tracks.addAll(index, newtracks)) {
                touch();
            }
        }
    }

    /**
     * Add the specified Track to the Playlist.
     * @param track the Track to add
     */
    public synchronized void add(Track track) {
        add(track, size());
    }

    /**
     * Add the specified Track to the Playlist.
     * @param track the Track to add
     * @param index the insertion position in the list
     */
    public synchronized void add(Track track, int index) {
        if (track==null) {
            throw new NullPointerException("Track is null");
        }
        if (index==tracks.size()) {
            tracks.add(track);
        } else {
            tracks.add(index, track);
        }
        touch();
    }

    public synchronized void clear() {
        tracks.clear();
        touch();
    }

    public synchronized Track remove(int index) {
        if (index < 0 || index >= tracks.size()) {
            throw new IllegalArgumentException("index "+index+" outside 0<=i<"+tracks.size());
        }
        Track track = tracks.remove(index);
        touch();
        return track;
    }

    public synchronized boolean removeTrack(Track track) {
        if (tracks.remove(track)) {
            touch();
            return true;
        } else {
            return false;
        }
    }

    public synchronized void removeAll(Playlist playlist) {
        if (tracks.removeAll(playlist.getTracks())) {
            touch();
        }
    }

    public synchronized void removeAll(BitSet bitset) {
        if (bitset.cardinality() > 0) {
            int diff = 0;
            for (int i=bitset.nextSetBit(0);i>=0;i=bitset.nextSetBit(i+1)) {
                tracks.remove(i - (diff++));
            }
            touch();
        }
    }


    public synchronized List<Track> getTracks() {
        if (db.getRevision() > lastdbrevision) {
            lastdbrevision = db.getRevision();
            Collection<Track> alltracks = db.getTracks();
            if (tracks.retainAll(alltracks)) {
                revision++;
            }
        }
        return Collections.unmodifiableList(tracks);
    }

    public String toString() {
        return "ManualPlaylist@"+Integer.toHexString(hashCode());
    }

    @Override
    public Playlist sort(final String sort, final String filter) {
        return new Playlist() {
            List<Track> filteredtracks;
            public synchronized List<Track>getTracks() {
                if (filteredtracks == null) {
                    filteredtracks = db.getTracks(sort, filter);
                    filteredtracks.retainAll(tracks);
                    filteredtracks = Collections.unmodifiableList(filteredtracks);
                }
                return filteredtracks;
            }
            public String getName() {
                return ManualPlaylistImpl.this.getName();
            }
            public void setName(String name) {
                ManualPlaylistImpl.this.setName(name);
            }
            public int getUniqueID() {
                return ManualPlaylistImpl.this.uniqueid;
            }
            public int getRevision() {
                return 0;
            }
            public int size() {
                return getTracks().size();
            }
            public Playlist sort(String sort, String filter) {
                return ManualPlaylistImpl.this.sort(sort, filter);
            }
            public String toString() {
                return ManualPlaylistImpl.this.toString()+"{"+sort+","+filter+"}";
            }
            public boolean equals(Object o) {
                return o == this;
            }
        };
    }

    public int size() {
        return getTracks().size();
    }

}
