package org.qtunes.db.spi.simple;

import org.qtunes.db.*;
import java.util.*;

/**
 * Simple in-memory indices which can be quickly derived from a Store.
 * Provides sorting and searching
 */
class SimpleIndices {
    
    private static final BitSet EMPTY = new BitSet();
    private static final Field[] SEARCHFIELDS = new Field[] { Field.Album, Field.Artist, Field.Title, Field.AlbumArtist, Field.Group, Field.Composer };
    private static final Field[] INDEXFIELDS = new Field[] { Field.Album, Field.Artist, Field.Title, Field.AlbumArtist, Field.Disc, Field.Track, Field.Year };
    private transient Map<String,Comparator<Integer>> map = new HashMap<String,Comparator<Integer>>();
    private transient Store store;

    private int[][] sort;
    private boolean dirty;
    private String[] keys;
    private EWAHCompressedBitmap[] vals;

    SimpleIndices(Store store) {
        this.store = store;
        map.put("+album", new SimpleComparator(new Field[] { Field.Album, Field.Disc, Field.Track, Field.Artist, Field.Title }, false));
        map.put("-album", new SimpleComparator(new Field[] { Field.Album, Field.Disc, Field.Track, Field.Artist, Field.Title }, true));
        map.put("+title", new SimpleComparator(new Field[] { Field.Title, Field.Artist, Field.Album, Field.Disc, Field.Track }, false));
        map.put("-title", new SimpleComparator(new Field[] { Field.Title, Field.Artist, Field.Album, Field.Disc, Field.Track }, true));
        map.put("+artist", new SimpleComparator(new Field[] { Field.Artist, Field.Album, Field.Disc, Field.Track, Field.Title }, false));
        map.put("-artist", new SimpleComparator(new Field[] { Field.Artist, Field.Album, Field.Disc, Field.Track, Field.Title }, true));
        map.put("+albumartist", new SimpleComparator(new Field[] { Field.AlbumArtist, Field.Album, Field.Disc, Field.Track, Field.Title }, false));
        map.put("-albumartist", new SimpleComparator(new Field[] { Field.AlbumArtist, Field.Album, Field.Disc, Field.Track, Field.Title }, true));
    }

    synchronized void setDirty() {
        dirty = true;
    }

    synchronized boolean isDirty() {
        return dirty;
    }

    synchronized void rebuild() {
        this.sort = buildSingleFieldIndices(store, INDEXFIELDS);

        Map<String,EWAHCompressedBitmap> all = new HashMap<String,EWAHCompressedBitmap>();
        BitSet tracks = store.getTracks();
        for (int i=tracks.nextSetBit(0);i>=0;i=tracks.nextSetBit(i+1)) {
            for (int j=0;j<SEARCHFIELDS.length;j++) {
                String val = (String)store.get(i, SEARCHFIELDS[j]);
                if (val != null) {
                    String[] values = NormalizingComparator.clean(val).split(" ");
                    for (int k=0;k<values.length;k++) {
                        val = values[k];
                        EWAHCompressedBitmap bs = all.get(val);
                        if (bs == null) {
                            all.put(val, bs = new EWAHCompressedBitmap());
                        }
                        bs.set(i);
                    }
                }
            }
        }
        keys = all.keySet().toArray(new String[0]);
        Arrays.sort(keys);
        vals = new EWAHCompressedBitmap[keys.length];
        for (int i=0;i<keys.length;i++) {
            vals[i] = all.get(keys[i]);
        }

        dirty = false;
    }

    public BitSet getFilteredTracks(String value) {
        return get(NormalizingComparator.clean(value).split(" "));
    }

    private BitSet get(String[] searchkeys) {
        EWAHCompressedBitmap bs = null;
        for (int i=0;i<searchkeys.length;i++) {
            String searchkey = searchkeys[i];
            int ix = Arrays.binarySearch(keys, searchkey);
            if (ix < 0) {
                ix = -ix - 1;
            }
            EWAHCompressedBitmap bs2 = null;
            while (ix < keys.length && keys[ix].startsWith(searchkey)) {
                bs2 = bs2 == null ? vals[ix] : bs2.or(vals[ix]);
                ix++;
            }
            bs = bs == null ? bs2 : bs.and(bs2);
        }
        return bs==null ? EMPTY : bs.getBitSet();
    }

    synchronized Comparator<Integer> getComparator(String key) {
        if (sort == null) {
            rebuild();
        }
        key = key.toLowerCase();
        Comparator<Integer> comp = map.get(key);
        if (comp == null) {
            throw new IllegalStateException("Unknown ordering "+key);
        }
        return comp;
    }

    @SuppressWarnings("unchecked")
    private int[][] buildSingleFieldIndices(Store store, Field[] fields) {
        BitSet tracks = store.getTracks();
        int[][] indices = new int[fields.length][tracks.length()];
        for (int i=0;i<fields.length;i++) {
            Field field = fields[i];
            Comparator comparator = field.getType()==String.class ? new NormalizingComparator(true, true) : null;
            SortedMap<Object,List<Integer>> sorted = new TreeMap<Object,List<Integer>>(comparator);
            for (int j=tracks.nextSetBit(0);j>=0;j=tracks.nextSetBit(j+1)) {
                Object val = store.get(j, field);
                if (val==null && field.getType()==String.class) {
                    val = "";
                }
                if (val!=null) {
                    List<Integer> list = sorted.get(val);
                    if (list==null) {
                        sorted.put(val, list = new ArrayList<Integer>());
                    }
                    list.add(Integer.valueOf(j));
                }
            }
            indices[i] = new int[tracks.length()];
            int p = 0;
            for (Iterator<List<Integer>> j=sorted.values().iterator();j.hasNext();) {
                List<Integer> list = j.next();
                p++;
                for (int k=0;k<list.size();k++) {
                    indices[i][list.get(k)] = p;
                }
            }
        }
        return indices;
    }

    private class SimpleComparator implements Comparator<Integer> {
        private final int[] fields;
        private final boolean reversefirst;
        public SimpleComparator(Field[] fields, boolean reversefirst) {
            this.fields = new int[fields.length];
            for (int i=0;i<fields.length;i++) {
                this.fields[i] = Arrays.asList(INDEXFIELDS).indexOf(fields[i]);
            }
            this.reversefirst = reversefirst;
        }

        public int compare(Integer t1, Integer t2) {
            int i1 = t1.intValue();
            int i2 = t2.intValue();
            for (int i=0;i<fields.length;i++) {
                int[] sortkeys = sort[fields[i]];
                // Bulletproofed to not crash while index is being rebuilt
                if (sortkeys != null) {
                    int s1 = i1 < sortkeys.length ? sortkeys[i1] : 0;
                    int s2 = i2 < sortkeys.length ? sortkeys[i2] : 0;
                    if (s1 != s2) {
                        return i==0 && reversefirst ? s2-s1 : s1-s2;
                    }
                }
            }
            return i1-i2;
        }
    }

}
