package org.qtunes.db.spi.simple;

import org.qtunes.core.*;
import org.qtunes.db.*;
import org.qtunes.ff.TrackReader;
import java.util.*;
import java.io.*;
import java.nio.*;
import java.nio.channels.*;
import javax.sound.sampled.*;

/**
 A Track is a single Track. It's created by the {@link Database#create} method, which in turn gets
 * its values from a {@link TrackReader}. A Track is made up of a number of {@link AbstractField}
 * classes, and it can be converted to audio sample by calling {@link #getAudioInputStream}. Some
 * types of {@link TrackReader} are also instances of {@link TrackWriter}, and if that's the case
 * the Track can be updated by calling {@link Database#updateTrack}.
 */
class TrackImpl implements Track {

    private int index;
    private DatabaseImpl db;
    private Map<Object,Object> properties;

    TrackImpl(int index, DatabaseImpl db) {
        if (index==0) throw new IllegalArgumentException();
        this.index = index;
        this.db = db;
    }

    public boolean equals(Object o) {
        return o instanceof TrackImpl && index==((TrackImpl)o).index && db==((TrackImpl)o).db;
    }

    public int hashCode() {
        return index ^ db.hashCode();
    }

    /**
     * Get the index of this track in the database
     */
    public int getIndex() {
        return index;
    }

    private synchronized Map<Object,Object> getProperties() {
        if (properties==null) {
            properties = new HashMap<Object,Object>();
        }
        return properties;
    }

    /**
     * Get the value of the specified Field. The return type
     * depends on the Field but should be a String, Integer, Boolean,
     * Float, Long, int[] or byte[], or it may be null. The other
     * get methods can be called to type the return value correctly.
     */
    @SuppressWarnings("unchecked")
    public <T> T get(AbstractField<T> field) {
        if (field instanceof Field) {
            T val = (T)db.getField(index, (Field)field);
            if (val==null) {
                if (((Field)field).getType()==Integer.class) {
                    val = (T)Integer.valueOf(0);
                } else if (((Field)field).getType()==Long.class) {
                    val = (T)Long.valueOf(0);
                } else if (((Field)field).getType()==Float.class) {
                    val = (T)Float.valueOf(0);
                } else if (((Field)field).getType()==Boolean.class) {
                    val = (T)Boolean.FALSE;
                }
            }
            return val;
        } else if (field instanceof DerivedField) {
            DerivedField dfield = (DerivedField)field;
            Object value = null;
            if (dfield.isCacheable()) {
                value = getProperties().get(dfield);
            }
            if (value==null) {
                value = dfield.get(this);
                if (dfield.isCacheable()) {
                    getProperties().put(dfield, value);
                }
            }
            return (T)value;
        } else {
            return null;
        }
    }

    /**
     * Get the File this Track was loaded from
     */
    public File getFile() {
        String field = get(Field.File);
        if (field == null) {
            return null;
        } else if (field.charAt(0)=='$') {
            field = db.getContext().getGlobalProperty("root") + field.substring(1);
        }
        return new File(field);
    }

    /**
     * Get the {@link AudioInputStream} for this track. Don't
     * forget to close the stream once you're done.
     */
    public AudioInputStream getAudioInputStream() throws IOException {
        // jlayer and jaad are both fine with their own formats, but either misidentify
        // other formats as their own, pass over them but leave them in a bad state
        // for the other AudioFileReaders, or probably both. We've already identified
        // the file type, so we get the TrackReader and the AudioInpuStream from that

        TrackReader reader = TrackReader.getTrackReader(db.getContext(), (String)get(Field.Type));
        try {
            return reader.getAudioInputStream(getFile());
        } catch (UnsupportedAudioFileException e) {
            throw new IOException(e);
        }
    }

    /**
     * Get a {@link ByteBuffer} containing the image. This maps
     * the file into memory, so the returned ByteBuffer should
     * be copied if the return value is not to be discarded
     * immediately.
     */
    public ByteBuffer getImageBuffer() throws IOException {
        final int off = get(Field.ImageOffset);
        final int len = get(Field.ImageLength);
        File file = getFile();
        if (off != 0 && len != 0 && file.exists()) {
            FileChannel in = new FileInputStream(file).getChannel();
            ByteBuffer buf = in.map(FileChannel.MapMode.READ_ONLY, off, len);
            in.close();
            return buf;
        } else {
            return null;
        }
    }

    /**
     * Get an Iterator that moves over all the {@link Field} objects
     * in this Track
     */
    public Iterator<Field<?>> getFields() {
        final BitSet fields = db.getFields(index);
        return new Iterator<Field<?>>() {
            int i = fields.nextSetBit(0);

            public boolean hasNext() {
                return i!=-1;
            }

            public Field<?> next() {
                int oldi = i;
                i = fields.nextSetBit(i+1);
                return Field.ALLFIELDS[oldi];
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }

    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("{ix:"+index);
        for (Iterator<Field<?>> i = getFields();i.hasNext();) {
            Field<?> field = i.next();
            Object o = get(field);
            if (o instanceof String) {
                o = "\""+o+"\"";
            }
            sb.append(", "+field.getName()+":"+o);
        }
        sb.append("}");
        return sb.toString();
    }

    public void update(FieldMap map) throws IOException {
        db.updateTrack(getIndex(), map);
    }

    public void delete() {
        db.deleteTrack(getIndex());
    }

}
