package org.qtunes.ff.spi.mp3;

import org.qtunes.core.util.*;
import org.qtunes.core.*;
import org.qtunes.ff.*;
import org.qtunes.db.*;
import java.io.*;
import java.util.*;
import java.nio.*;
import java.nio.channels.*;
import javax.sound.sampled.*;

public class MP3Handler extends TrackWriter {

    private ServiceContext context;
    private static final Map tags = new HashMap();
    private static final Set unknowns = new HashSet();
    private final TagMapper tagmapper = new TagMapper(this);
    private boolean gainitunes = true, gainrva = false, gainreplaygain = false;

    public String getType() {
        return "mp3";
    }

    public boolean matches(File file) {
        return file.getName().toLowerCase().endsWith(".mp3");
    }

    public void setGainStorage(String type, boolean on) {
        if ("itunes".equalsIgnoreCase(type)) {
            gainitunes = on;
        } else if ("rva".equalsIgnoreCase(type)) {
            gainrva = on;
        } else if ("replaygain".equalsIgnoreCase(type)) {
            gainreplaygain = on;
        }
    }

    public boolean getGainStorage(String type) {
        if (getContext() != null) {
            return "true".equals(getContext().getProperty("gain."+type));
        } else {
            if ("itunes".equalsIgnoreCase(type)) {
                return gainitunes;
            } else if ("rva".equalsIgnoreCase(type)) {
                return gainrva;
            } else if ("replaygain".equalsIgnoreCase(type)) {
                return gainreplaygain;
            } else {
                return false;
            }
        }
    }

    //-----------------------------------------------------------------------------------

    public FieldMap read(File file) throws IOException {
        File absfile = FileUtils.getAbsoluteFile(file);
        FileChannel in = new FileInputStream(absfile).getChannel();

        final FieldMap fields = new FieldMap();
        final String filename = file.getPath();
        fields.put(Field.Type, "mp3");
        fields.put(Field.Codec, "mp3");
        fields.put(Field.File, filename);
        fields.put(Field.FileSize, Integer.valueOf((int)absfile.length()));
        fields.put(Field.LastModified, Integer.valueOf((int)(absfile.lastModified()/1000)));

        long rangestart = 0;
        long rangeend;
        TagHandler taghandler = new TagHandler() {
            public void beginBlock(int version) {
                if (version==2) {
                    fields.put(Field.ID3v22, true);
                } else if (version==3) {
                    fields.put(Field.ID3v24, true);
                } else if (version==4) {
                    fields.put(Field.ID3v23, true);
                }
            }
            public void endBlock(int offset) { }
            public void tag(Tag tag) throws IOException {
                tagmapper.setFields(tag, fields);
            }
        };

        if (!readID3v2(in, filename, taghandler)) {
            rangestart = 0;
            if (readID3v1(in, fields, filename)) {
                rangeend = in.size() - 128;
            } else {
                rangeend = in.size();
            }
        } else {
            rangestart = in.position();
            if (hasID3v1(in)) {
                fields.put(Field.ID3v1, true);
                rangeend = in.size() - 128;
            } else {
                rangeend = in.size();
            }
        }

        in.position(rangestart);
        readBitrate(in, fields, filename);
        in.close();

        // Some tracks have TPE2 (album artist) but not TPE1 (artist) -
        // in that case we also use albumartist as artist
        if (fields.containsKey(Field.AlbumArtist) && !fields.containsKey(Field.Artist)) {
            fields.putUntyped(Field.Artist, fields.get(Field.AlbumArtist));
        }
        fields.put(Field.AudioByteRange, new Long((rangestart<<32) | (int)(rangeend-rangestart)));
        return fields;
    }

    private static boolean readBuffer(FileChannel in, ByteBuffer buf) throws IOException {
        buf.rewind();
        while (buf.remaining() > 0) {
            if (in.read(buf) < 0) {
                throw new EOFException("EOF at "+buf.remaining()+", expected "+buf.capacity());
            }
        }
        buf.flip();
        return true;
    }

    private boolean readID3v2(FileChannel channel, String filename, TagHandler handler) throws IOException {
        String id3version = null;
        ByteBuffer headbuf = ByteBuffer.allocate(10);

	channel.position(0);
        long lastheader;
	while ((lastheader = channel.position())>=0 && readBuffer(channel, headbuf) && headbuf.get()=='I' && headbuf.get()=='D' && headbuf.get()=='3') {
            String version = "2."+headbuf.get()+"."+headbuf.get();
	    int id3flags = headbuf.get();
	    int id3v2len =  Tag.readSyncInt(headbuf);
            long start = channel.position();
            ByteBuffer buf = ByteBuffer.allocate(id3v2len);
            readBuffer(channel, buf);

	    if ((id3flags&0x40) == 0x40) {        // Extended Header (not used)
		int size = buf.getInt();
                buf.position(buf.position() + size);
	    }
            int id3v2v = 0;
            if ("2.2.0".equals(version)) {
                handler.beginBlock(id3v2v = 2);
            } else if ("2.3.0".equals(version)) {
                handler.beginBlock(id3v2v = 3);
            } else if ("2.4.0".equals(version)) {
                handler.beginBlock(id3v2v = 4);
            }

	    if (id3version==null || version.compareTo(id3version) > 0) { // Later versions over earlier
		id3version = version;
                Tag tag;
                while ((tag = Tag.read(buf, id3v2v, start))!=null) {
                    try {
//                        System.out.println("Read tag="+tag+" at "+tag.getFileOffset());
                        if (tag.getVersion() == id3v2v) {
                            handler.tag(tag);
                        } else {
                            log("WARNING: Read v"+tag.getVersion()+" tag \""+tag.getName()+"\" in v"+id3v2v+" table", null);
                        }
                    } catch (Exception e) {
                        String message;
                        if (e instanceof IOException) {
                            message = "Invalid "+tag.getName()+" in \""+filename+"\": ("+e.getMessage()+")";
                        } else {
                            message = "Invalid "+tag.getName()+" in \""+filename+"\": "+e.getClass().getName()+" ("+e.getMessage()+")";
                        }
                        log (message, e);
                    }
                }
            }
	}
        channel.position(lastheader);
        return id3version != null;
    }

    private static boolean hasID3v1(FileChannel in) throws IOException {
        ByteBuffer buf = ByteBuffer.allocate(128);
        long pos = in.position();
        in.position(in.size() - 128);
        readBuffer(in, buf);
        in.position(pos);

        return buf.get()=='T' && buf.get()=='A' && buf.get()=='G';
    }

    /**
     * Set Duration, SampleRate, BitRate, Mono, VBR
     */
    private void readBitrate(FileChannel channel, FieldMap fields, String filename) throws IOException {
        long start = channel.position();
        int numaudiobytes = (int)(channel.size() - start);
        ByteBuffer buf = ByteBuffer.allocateDirect(8192);       // Direct - we're mostly seeking
        readBuffer(channel, buf);                               // and not accessing buffer as array.
        Frame frame = readFrame(buf, start);
        if (frame != null) {
            fields.put(Field.SampleRate, Integer.valueOf(frame.samplerate));
            if (frame.mode == Frame.MONO) {
                fields.put(Field.Mono, true);
            }

            // First check for "Xing" header
            int xingpos = (frame.version==Frame.MPEG1 ? (frame.mode==Frame.MONO ? 21 : 36) : (frame.mode==Frame.MONO ? 23 : 21));
            int numframes = -1;
            boolean vbr = false;
            if (buf.getInt(xingpos) == 0x58696e67) {
                int head = buf.getInt(xingpos + 4);
                vbr = true;
                if ((head&0x00000001)==1) {
                    buf.position(xingpos + 8);
                    numframes = buf.getInt();
                    if ((head&0x00000002)==2) {
                        numaudiobytes = buf.getInt();
                    }
                    float tpf = (Frame.VBRTABLE[frame.layer] / frame.samplerate) / (frame.version==Frame.MPEG2 || frame.version==Frame.MPEG25 ? 2 : 1) * 1000;
                    int time = (int)(tpf * numframes);
                    int bitrate = (int)(numaudiobytes * 8l / time);
                    fields.put(Field.Duration, Integer.valueOf(time));
                    fields.put(Field.BitRate, Integer.valueOf(bitrate));
                }
            }
            if (numframes < 0) {
                // No Xing header. First check the 8k buffer we have for different bitrates,
                // if not then bring in another 8k halfway through song and check there too.
                Frame frame2;
                while (!vbr  && (frame2=readFrame(buf, start))!=null) {
                    vbr = frame.bitrate != frame2.bitrate;
                }
                if (!vbr && numaudiobytes > buf.capacity() * 2) {
                    channel.position(start + numaudiobytes/2);
                    buf.clear();
                    readBuffer(channel, buf);
                    while (!vbr  && (frame2=readFrame(buf, start))!=null) {
                        vbr = frame.bitrate != frame2.bitrate;
                    }
                }

                float msperframe = (frame.layer==Frame.LAYER1 ? 384f : 1152f) / frame.samplerate * 1000;
                if (vbr) {
                    // VBR but no header (LAME, or theoretically Xing with no "numframes" field).
                    // Scan file, counting frames (which are fixed duration) and calc average bitrate.
                    buf = ByteBuffer.allocateDirect(65536);
                    channel.position(start);
                    numframes = 0;
                    long bitrate = 0;
                    boolean readany = true;
                    while (readany && channel.read(buf) > 0) {
                        buf.flip();
                        readany = false;
                        while ((frame = readFrame(buf, start)) != null) {
                            numframes++;
                            readany = true;
                            bitrate += frame.bitrate;
                        }
                        channel.position(channel.position() - buf.remaining());
                        start += buf.position();
                        buf.clear();
                    }
                    bitrate /= numframes;
                    fields.put(Field.BitRate, Integer.valueOf((int)bitrate));
                } else {
                    // Fixed rate - num frames determined from remaining file length
                    numframes = numaudiobytes / frame.framelength;
                    fields.put(Field.BitRate, Integer.valueOf(frame.bitrate));
                }
                int time = (int)(numframes * msperframe);
                fields.put(Field.Duration, Integer.valueOf(time));
            }
            if (vbr) {
                fields.put(Field.VBR, Boolean.TRUE);
            }
        }
    }

    private static class Frame {
        static final int MPEG1=3, MPEG2=2, MPEG25=0, LAYER1=3, LAYER2=2, LAYER3=1, MONO=3, STEREO=0, JSTEREO=1, DUAL=2;
	static final int[][][] BITRATES = {
            { null, {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, -1}, {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, -1}, {0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256, -1} },
            { null, null, null, null },
            { null, {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, -1}, {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, -1}, {0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256, -1} },
            { null, {0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, -1}, {0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384, -1}, {0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448, -1} }
        };
        static final int[][] SAMPLERATES = { {11025, 12000, 8000, 1}, {-1, -1, -1, -1}, {22050, 24000, 16000, 1}, {44100, 48000, 32000, 1} };
	static final float[] VBRTABLE = {-1, 1152, 1152, 384 };

        long framestart;
        int version, layer, padding, samplerate, bitrate, mode, framelength;

        public String toString() {
            return "[F: v="+version+" l="+layer+" pos="+framestart+":"+framelength+" br="+bitrate+" sr="+samplerate+"]";
        }
    }

    private Frame readFrame(ByteBuffer in, long startpos) {
        int c, start;
        while (in.remaining() >= 4) {
            start = in.position();
            c = in.get() & 0xFF;
            if (c == 0xFF) {
                c = in.get();
                if ((c&0xE0) == 0xE0) {
                    int head = 0xFF000000 | ((c&0xFF)<<16) | (in.getShort()&0xFFFF);
                    int version = (head >>> 19) & 3;
                    int layer = (head >>> 17) & 3;
                    int bitrateindex = (head >>> 12) & 15;
                    int samplerateindex = (head >>> 10) & 3;
                    int bitrate, samplerate;
                    if (bitrateindex != 0x0F && samplerateindex != 0x03 && layer != 0 && version != 1 && (bitrate=Frame.BITRATES[version][layer][bitrateindex]) >= 0 && (samplerate = Frame.SAMPLERATES[version][samplerateindex]) >= 0) {
                        int padding = (head >>> 9) & 1;
                        int framelength;
                        if (layer == Frame.LAYER1) {
                            framelength = (12 * bitrate * 1000 / samplerate + padding) << 2;
                        } else {
                            framelength = (version==Frame.MPEG2 || version==Frame.MPEG25 ? 72 : 144) * bitrate * 1000 / samplerate + padding;
                        }
                        if (framelength >= 21) {    // Min
                            if (in.position() + framelength > in.limit()) {
                                in.position(start);
                                return null;
                            }
                            Frame frame = new Frame();
                            frame.version = version;
                            frame.layer = layer;
                            frame.bitrate = bitrate;
                            frame.samplerate = samplerate;
                            frame.mode = (head >>> 6) & 3;
                            frame.framestart = start + startpos;
                            frame.framelength = framelength;
                            in.position(in.position() + framelength - 4);
                            return frame;
                        }
                    }
                }
                in.position(start+1);
            }
        }
        return null;
    }

    private boolean readID3v1(FileChannel in, FieldMap fields, String filename) throws IOException {
        byte[] id3 = new byte[128];
        ByteBuffer buf = ByteBuffer.wrap(id3);
	in.position(in.size() - 128);
        readBuffer(in, buf);

        String title, artist, album, year, track, comment, genre;
        String val;
        if (id3[0]=='T' && id3[1]=='A' && id3[2]=='G') {
            fields.put(Field.ID3v1, true);
            if ((val = new String(id3, 3, 30, "ISO-8859-1").trim()).length() > 0) {
                fields.put(Field.Title, val);
            }
            if ((val = new String(id3, 33, 30, "ISO-8859-1").trim()).length() > 0) {
                fields.put(Field.Artist, val);
            }
            if ((val = new String(id3, 63, 30, "ISO-8859-1").trim()).length() > 0) {
                fields.put(Field.Album, val);
            }
            if ((val = new String(id3, 93, 4, "ISO-8859-1").trim()).length() > 0) {
                try {
                    fields.put(Field.Year, Integer.valueOf(val));
                } catch (NumberFormatException e) { }
            }
            if (id3[126]!=0) {
                fields.put(Field.Track, new Integer(id3[126] & 0xFF));
                val = new String(id3, 97, 28, "ISO-8859-1").trim();
            } else {
                val = new String(id3, 97, 30, "ISO-8859-1").trim();
            }
            if (val.length() > 0) {
                fields.put(Field.Comment, val);
            }
            if (id3[127] < 126 && id3[127] >= 0) {
                fields.put(Field.Genre, MP3GENRES[id3[127]]);
            }
            return true;
        } else {
            return false;
        }
    }

    @SuppressWarnings("unchecked")
    public void write(Track track, FieldMap changes) throws IOException {
        File infile = track.getFile();

        // Check for first time tagging
        if (!track.get(Field.ID3v22) && !track.get(Field.ID3v23) && !track.get(Field.ID3v24)) {
            FieldMap newchanges = new FieldMap();
            for (Iterator<Field<?>> i = track.getFields();i.hasNext();) {
                Field<?> field = i.next();
                Object value = track.get(field);
                newchanges.putUntyped(field, value);
            }
            newchanges.putAll(changes);
            changes = newchanges;
        }
        write(infile, changes);
    }

    @SuppressWarnings("unchecked")
    public boolean write(final File infile, final FieldMap fields) throws IOException {
        File absinfile = FileUtils.getAbsoluteFile(infile);
        long lastmodified = fields.containsKey(Field.LastModified) ? absinfile.lastModified() : 0;
        FileChannel inchannel = null, outchannel = null;
        File outfile = null;
        try {
            inchannel = new FileInputStream(absinfile).getChannel();

            final List<Tag> tags = new ArrayList<Tag>();
            final int[] maxversion = new int[1];
            TagHandler handler = new TagHandler() {
                public void beginBlock(int version) {}
                public void endBlock(int fileoffset) {}
                public void tag(Tag tag) throws IOException {
                    tags.add(tag);
                    maxversion[0] = Math.max(maxversion[0], tag.getVersion());
                }
            };
            readID3v2(inchannel, infile.getPath(), handler);
            int audiostart = (int)inchannel.position();
            boolean stripid3v1 = hasID3v1(inchannel) && !fields.containsKey(Field.ID3v1);

            List<Tag> oldtags = new ArrayList(tags);

            for (int i=0;i<tags.size();i++) {
                boolean invalid = false;
                String name = tags.get(i).getName();
                if (name.length() != (maxversion[0] == 2 ? 3 : 4)) {
                    invalid = true;
                }
                for (int j=0;j<name.length();j++) {
                    char c = name.charAt(j);
                    if ("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789".indexOf(c) < 0) {
                        invalid = true;
                    }
                }
                if (invalid) {
                    log ("WARNING: Removing invalid tag \""+tags.get(i)+"\" at "+i, null);
                    tags.remove(i--);
                }
            }

            for (Iterator<Map.Entry<Field<?>,Object>> i = fields.entrySet().iterator();i.hasNext();) {
                Map.Entry<Field<?>,Object> e = i.next();
                Field<?> field = e.getKey();
                Object value = e.getValue();
                tagmapper.setTags(field, value, tags, maxversion[0]);
//                System.out.println("Setting tag: field="+field+" Value="+value+" v="+maxversion[0]);
            }
            if (tags.equals(oldtags) && !stripid3v1) {
                return false;
            } else {
                stripid3v1 = true;      // If we're changing ID3v2, remove ID3v1
            }

            int size = 10;
            for (int i=0;i<tags.size();i++) {
                size += tags.get(i).length();
            }
            if (size < audiostart) {
                size = audiostart;
            }

            ByteBuffer header = ByteBuffer.allocate(20);
            header.put((byte)'I');
            header.put((byte)'D');
            header.put((byte)'3');
            header.put((byte)maxversion[0]);
            header.put((byte)0);
            header.put((byte)0);  // flags
            Tag.writeSyncInt(size - 10, header);
            header.flip();

            if (size > audiostart) {
                outfile = File.createTempFile(absinfile.getName(), null, absinfile.getParentFile());
                outchannel = new FileOutputStream(outfile).getChannel();
            } else {
                inchannel.close();
                inchannel = null;
                RandomAccessFile file = new RandomAccessFile(absinfile, "rw");
                outchannel = file.getChannel();
                if (stripid3v1) {
                    file.setLength(file.length() - 128);
                }
            }

            outchannel.write(header);
            for (int i=0;i<tags.size();i++) {
                ByteBuffer buf = tags.get(i).getAsByteBuffer();
//                System.out.println("Writing "+buf.limit()+" of tag "+tags.get(i));
                outchannel.write(buf);
            }

            if (size > audiostart) {
//              System.out.println("Writing remainder from "+audiostart);
                int blocksize = 67076096;   // Max block size for Windows
                long insize = inchannel.size() - (stripid3v1 ? 128 : 0);
                long position = audiostart;
                while (position < insize) {
                   position += inchannel.transferTo(position, blocksize, outchannel);
                }
                inchannel.close();
                inchannel = null;
                outchannel.close();
                outchannel = null;
                if (!outfile.renameTo(absinfile)) {
                    throw new IOException("Unable to rename \""+outfile+"\" to \""+absinfile+"\"");
                }
                if (lastmodified != 0) {
                    absinfile.setLastModified(lastmodified);
                }
                outfile = null;
            } else {
                ByteBuffer buf = ByteBuffer.allocate(1024);
                while (outchannel.position() < audiostart) {
                    buf.limit((int)Math.min(buf.capacity(), audiostart - outchannel.position()));
//                    System.out.println("Writing "+buf.limit()+" to "+outchannel.position());
                    outchannel.write(buf);
                    buf.clear();
                }
                outfile = null;
            }
        } finally {
            if (inchannel!=null) {
                inchannel.close();
            }
            if (outchannel!=null) {
                outchannel.close();
            }
            if (outfile!=null) {
                outfile.delete();
            }
        }
        return true;
    }

    private static interface TagHandler {
        public void beginBlock(int version);
        public void endBlock(int fileoffset);
        public void tag(Tag tag) throws IOException;
    }

    public void startService(ServiceContext context) {
        this.context = context;
    }

    public void stopService(ServiceContext context) {
    }

    public ServiceContext getContext() {
        return context;
    }

    public Map<String,Object> reportState() {
        Map<String,Object> map = new LinkedHashMap<String,Object>();
        map.put("gain.itunes", "true".equals(context.getProperty("gain.itunes")));
        map.put("gain.rva", "true".equals(context.getProperty("gain.rva")));
        map.put("gain.replaygain", "true".equals(context.getProperty("gain.replaygain")));
        return map;
    }

    private void log(String message, Exception e) {
        if (context != null) {
            context.debug(message);
        } else {        // Only when we're running this class directly, for debugging
            System.out.println(message);
            if (e != null) {
                e.printStackTrace();
            }
        }
    }

    public static void main(String[] args) throws Exception {
        MP3Handler handler = new MP3Handler();
        for (int i=0;i<args.length;i++) {
            FieldMap map = handler.read(new File(args[i]));
            System.out.println(map);
            handler.getAudioInputStream(new File(args[i]));
        }
    }

    public AudioInputStream getAudioInputStream(File file) throws UnsupportedAudioFileException, IOException {
       return JLInputStream.getInputStream(file);
    }

}
