package org.qtunes.ff.spi.mp3;

import org.qtunes.core.*;
import org.qtunes.db.*;
import java.io.*;
import java.util.*;
import java.nio.*;

class Tag {

    private static final Properties TAGCLASSES;
    private String name;
    private int flags, version, dataoffset;
    private long fileoffset;
    private ByteBuffer data;

    static {
        try {
            Object[] required = { TagComment.class, TagGenre.class, TagImage.class, TagInteger.class, TagNumber.class, TagText.class, TagPosition.class, TagTimestamp.class, TagTypedText.class, TagUFID.class };       // We need these, make sure they're compiled
            TAGCLASSES = new Properties();
            InputStream in = Tag.class.getResourceAsStream("tagclasses.properties");
            TAGCLASSES.load(in);
            in.close();
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    Tag() {
    }

    Tag(String name, int version, int flags, byte[] data) {
        this.version = version;
        this.flags = flags;
        this.name = name;
        if (data!=null) {
            this.data = ByteBuffer.allocate(data.length);
            this.data.put(data);
            this.data.rewind();
        }
    }

    final static Tag read(ByteBuffer in, int version, long offset) throws IOException {
        int namelen = version==2 ? 3 : 4;
        int start = in.position();
        String name = in.remaining() > namelen + 3 ? readFixedString(in, version==2 ? 3 : 4) : null;
        if (name == null || "\0\0\0\0".equals(name) || "\0\0\0".equals(name)) {
            return null;
        } else {
            boolean bad3 = false;
            if (version > 2 && name.charAt(3) == 0) {
                name = name.substring(0, 3);
                bad3 = true;
            }
            int size, flags;
            if (version == 4) {
                size = readSyncInt(in);
                flags = ((in.get()&0xFF)<<8) + (in.get()&0xFF);
            } else if (version == 3) {
                size = in.getInt();
                flags = ((in.get()&0xFF)<<8) + (in.get()&0xFF);
                if (bad3) {
                    in.get();
                    size--;
                    version = 2;
                }
            } else {
                size = ((in.get()&0xFF)<<16) + ((in.get()&0xFF)<<8) + (in.get()&0xFF);
                flags = 0;
            }
            if (size > in.limit() || size < 0) {      // Catch bad data
                return null;
            }
            String type = (String)TAGCLASSES.get(name);
            Tag tag;
            if (type==null && name.charAt(0)=='W') {
                type = "TagText";
            }
            if (type==null) {
                tag = new Tag();
            } else {
                try {
                    String classname = Tag.class.getName();
                    classname = classname.substring(0, classname.lastIndexOf(".")+1)+type;
                    tag = (Tag)Class.forName(classname).newInstance();
                } catch (Throwable e) {
                    throw new RuntimeException(e);
                }
            }
            tag.name = name;
            tag.fileoffset = offset + start;
            tag.version = version;
            tag.flags = flags;
            int oldlimit = in.limit();
            in.limit(in.position() + size);
            tag.dataoffset = in.position() - start;
            tag.data = (ByteBuffer)in.slice();
            in.limit(oldlimit).position(in.position() + size);
            return tag;
        }
    }

    final static void writeSyncInt(int i, ByteBuffer out) {
        out.put((byte)((i>>21)&0x7f));
        out.put((byte)((i>>14)&0x7f));
        out.put((byte)((i>>7)&0x7f));
        out.put((byte)(i&0x7f));
    }

    final static int getStringEncoding(int version, String value) {
        int encoding;
        if (version == 4) {
            encoding = 3;
        } else {
            encoding = 0;
            for (int i=0;encoding==0 && i<value.length();i++) {
                if (value.charAt(i) > 127 || value.charAt(i) < 32) {
                    encoding = 2;
                }
            }
        }
        return encoding;
    }

    boolean isInvalid() {
        return false;
    }

    final static void writeString(int encoding, String value, OutputStream out) throws IOException {
        OutputStreamWriter w = new OutputStreamWriter(out, encoding==3 ? "UTF-8" : encoding==0 ? "ISO-8859-1" : "UTF-16BE");
        if (encoding==1) {
            w.write((char)0xFEFF);
        }
        w.write(value);
        w.flush();
    }

    final static void writeStringTerminator(int encoding, OutputStream out) throws IOException {
        out.write(0);
        if (encoding == 1) {
            out.write(0);
        }
    }

    ByteBuffer getAsByteBuffer() {
        ByteBuffer buf = ByteBuffer.allocate(length());
        buf.put((byte)name.charAt(0));
        buf.put((byte)name.charAt(1));
        buf.put((byte)name.charAt(2));
        int size = data.limit();
        if (version==2) {
            buf.put((byte)(size>>16));
            buf.put((byte)(size>>8));
            buf.put((byte)size);
        } else {
            buf.put((byte)name.charAt(3));
            if (version==3) {
                buf.putInt(size);
            } else {
                writeSyncInt(size, buf);
            }
            buf.putShort((short)flags);
        }
        data.rewind();
        buf.put(data);
        buf.flip();
        return buf;
    }

    public boolean equals(Object o) {
        if (o.getClass() == getClass()) {
            Tag tag = (Tag)o;
            if (name.equals(tag.name) && flags==tag.flags && version==tag.version && data.limit()==tag.data.limit()) {
                for (int i=0;i<data.limit();i++) {
                    if (data.get(i) != tag.data.get(i)) {
                        return false;
                    }
                }
                return true;
            }
        }
        return false;
    }

    /**
     * Return the offset in the file of the start of this tag
     */
    final long getFileOffset() {
        return fileoffset;
    }

    /**
     * Return the offset from the start of this tag the source of the start of this tag's content.
     */
    protected final int getContentOffset() {
        return dataoffset;
    }

    /**
     * Return the content of this tag (not including the header or size).
     */
    final ByteBuffer getContent() {
        return data;
    }

    /**
     * Return the length of the overall tag, including the header 
     */
    final int length() {
        return data.capacity() + (version==2 ? 6 : 10);
    }

    /**
     * Return the 3 or 4-letter name of the tag
     */
    final String getName() {
        return name;
    }

    /**
     * Return the flags for this tag (unused but part of the spec)
     */
    final int getFlags() {
        return flags;
    }

    /**
     * Return the version of the tag - 2, 3 or 4 for 2.2, 2.3 or 2.4
     */
    final int getVersion() {
        return version;
    }

    /**
     * Read a String in the specified encoding from the ByteBuffer
     * @param encoding the encoding, as specified in the ID32.4 spec. This
     * is saved as the first char of the buffer, so if it's outside the range
     * 0-3 we assume the String does not include an encoding, and the encoding
     * byte will be treated as part of the string.
     */
    final static String readEncodedString(ByteBuffer in, byte encoding) throws IOException {
        StringBuilder sb = new StringBuilder();
        if (encoding == 0 || encoding > 3) {    // ISO-8859-1
            if (encoding > 3) {         // Catch case where no encoding specified
                sb.append((char)(encoding&0xFF));
            }
            byte b;
            while (in.remaining() > 0 && (b=in.get())!=0) {
                sb.append((char)(b&0xFF));
            }
        } else if (encoding == 1) {             // UTF-16 with possible BOM
            if (in.remaining() >= 2) {
                char c = (char)(((in.get()&0xFF)<<8) + (in.get()&0xFF));
                if (c==0xFFFE) {
                    in.order(ByteOrder.LITTLE_ENDIAN);
                } else if (c==0xFEFF) {
                    in.order(ByteOrder.BIG_ENDIAN);
                } else {
                    in.order(ByteOrder.BIG_ENDIAN);
                    sb.append(c);
                }
                while (in.remaining() > 1 && (c=in.getChar())!=0) {
                    sb.append(c);
                }
            }
        } else if (encoding == 2) {             // UTF-16BE
            char c;
            in.order(ByteOrder.BIG_ENDIAN);
            while (in.remaining() > 1 && (c=in.getChar())!=0) {
                sb.append(c);
            }
        } else if (encoding == 3) {             // UTF-8
            byte b;
            char c;
            while (in.remaining() > 0 && (b=in.get())!=0) {
                if (b <= 0x7F) {
                    c = (char)(b&0xFF);
                } else if (b <= 0xC1) {
                    throw new IOException("Invalid UTF8 sequence");
                } else if (b <= 0xDF && in.remaining() > 0) {
                    c = (char)(((b&0x1F)<<6) + (in.get()&0x3F));
                } else if (b <= 0xEF && in.remaining() > 1) {
                    c = (char)(((b&0x0F)<<12) + ((in.get()&0x3F)<<6) + (in.get()&0x3F));
                } else if (b <= 0xF4 && in.remaining() > 2) {
                    c = (char)(((b&0x07)<<18) + ((in.get()&0x3F)<<12) + ((in.get()&0x3F)<<6) + (in.get()&0x3F));
                } else {
                    // Should read 4,5,6 bytes sequences here?
                    throw new IOException("Invalid UTF8 sequence");
                }
                sb.append(c);
            }
        } else {
            throw new IOException("Invalid encoding "+encoding);
        }
        return sb.toString();
    }

    /**
     * Read a fixed-length String from the ByteBuffer in ISO-8859-1
     */
    final static String readFixedString(ByteBuffer in, int len) throws IOException {
        if (in.remaining() < len) {
            throw new EOFException("EOF reading "+len+" byte String");
        }
        StringBuilder sb = new StringBuilder(len);
        for (int i=0;i<len;i++) {
            sb.append((char)(in.get()&0xFF));
        }
        return sb.toString();
    }

    /**
     * Read a sync-int from the Stream, as described in ID3v2.4 spec
     */
    final static int readSyncInt(ByteBuffer in) {
        return ((in.get()&0x7F)<<21) + ((in.get()&0x7F)<<14) + ((in.get()&0x7F)<<7) + (in.get()&0x7F);
    }

    public String toString() {
        return "["+getName()+" data="+getContent().capacity()+" bytes]";
    }


}
