package org.qtunes.ff.spi.mp4;

import org.qtunes.core.util.*;
import org.qtunes.core.*;
import org.qtunes.ff.*;
import org.qtunes.db.*;
import java.util.*;
import java.io.*;
import java.nio.*;
import java.nio.channels.*;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.spi.AudioFileReader;
import javax.sound.sampled.UnsupportedAudioFileException;

/**
 * A parser to load MP4 objects. Doesn't decode audio, but can extract metadata
 * from the MP4 object. Uses NIO and skips over blocks that aren't required so
 * should be very quick and memory efficient.
 */
public class MP4Handler extends TrackReader {

    private ServiceContext context;

    final Box NULL = new Box(0, null, 0, 0, 0, null);
    static int[] containers = new int[] {
        0x6d6f6f76,    // moov
        0x7472616b,    // trak
        0x65647473,    // edts
        0x6d646961,    // mdia
        0x6d696e66,    // minf
        0x64696e66,    // dinf
        0x7374626c,    // stbl
        0x73747364,    // stsd
        0x6d703461,    // mp4a
        0x6d766578,    // mvex
        0x6d6f6f66,    // moof
        0x74726166,    // traf
        0x6d667261,    // mfra
        0x75647461,    // udta
        0x6970726f,    // ipro
        0x73696e66,    // sinf
        0x6d657461,    // meta
        0x696c7374,    // ilst
        0x2d2d2d2d,    // ----
        0xa9616c62,    // ?alb
        0xa9617274,    // ?art
        0x61415254,    // aART
        0xa9636d74,    // ?cmt
        0xa9646179,    // ?day
        0xa96e616d,    // ?nam
        0xa967656e,    // ?gen
        0x676e7265,    // gnre
        0x74726b6e,    // trkn
        0x6469736b,    // disk
        0xa9777274,    // ?wrt
        0xa9746f6f,    // ?too
        0x746d706f,    // tmpo
        0x63707274,    // cprt
        0x6370696c,    // cpil
        0x636f7672,    // covr
        0x72746e67,    // rtng
        0xa9677270,    // ?grp
        0x7374696b,    // stik
        0x70637374,    // pcst
        0x63617467,    // catg
        0x6b657977,    // keyw
        0x7075726c,    // purl
        0x65676964,    // egid
        0x64657363,    // desc
        0xa96c7972,    // ?lyr
        0x74766e6e,    // tvnn
        0x74767368,    // tvsh
        0x7476656e,    // tven
        0x7476736e,    // tvsn
        0x74766573,    // tves
        0x70757264,    // purd
        0x70676170,    // pgap
    };

    static Map<String,Field> tag2field = new HashMap<String,Field>();
    static {
        tag2field.put("moov.udta.meta.ilst.\u00A9alb", Field.Album);
        tag2field.put("moov.udta.meta.ilst.aART", Field.AlbumArtist);
        tag2field.put("moov.udta.meta.ilst.\u00A9ART", Field.AlbumArtist);
        tag2field.put("moov.udta.meta.ilst.\u00A9cmt", Field.Comment);
        tag2field.put("moov.udta.meta.ilst.\u00A9day", Field.Year);
        tag2field.put("moov.udta.meta.ilst.\u00A9nam", Field.Title);
        tag2field.put("moov.udta.meta.ilst.\u00A9wrt", Field.Composer);
        tag2field.put("moov.udta.meta.ilst.\u00A9too", Field.Encoder);
        tag2field.put("moov.udta.meta.ilst.cprt", Field.Copyright);
        tag2field.put("moov.udta.meta.ilst.\u00A9grp", Field.Group);
        tag2field.put("moov.udta.meta.ilst.\u00A9gen", Field.Genre);
        tag2field.put("moov.udta.meta.ilst.gnre", Field.Genre);
        tag2field.put("moov.udta.meta.ilst.trkn", Field.Track);
        tag2field.put("moov.udta.meta.ilst.disk", Field.Disc);
        tag2field.put("moov.udta.meta.ilst.tmpo", Field.BPM);
        tag2field.put("moov.udta.meta.ilst.cpil", Field.Compilation);
        tag2field.put("moov.udta.meta.ilst.covr", Field.ImageOffset);
        tag2field.put("moov.udta.meta.ilst.pgap", Field.Gapless);
        tag2field.put("moov.udta.meta.ilst.com.apple.iTunes.iTunNORM", Field.Gain);
        tag2field.put("moov.udta.meta.ilst.com.apple.iTunes.iTunes_CDDB_1", Field.CDDB);
    }


    public String getType() {
        return "mp4";
    }

    public boolean matches(File file) {
        String name = file.getName();
        return name.endsWith(".mp4") || name.endsWith(".m4a") || name.endsWith(".m4b");
    }

    public FieldMap read(File file) throws IOException {
        File absfile = FileUtils.getAbsoluteFile(file);
        FileChannel in = new FileInputStream(absfile).getChannel();

        final FieldMap fields = new FieldMap();
        final String filename = file.getPath();
        fields.put(Field.Type, "mp4");
        fields.put(Field.File, filename);
        fields.put(Field.FileSize, new Integer((int)absfile.length()));
        fields.put(Field.LastModified, new Integer((int)(absfile.lastModified()/1000)));
        parse(getBox(in, 0, in.size(), null), null, fields);
        return fields;
    }

    private static String typeString(int i) {
        char[] c = new char[4];
        c[0] = (char)((i>>24)&0xFF);
        c[1] = (char)((i>>16)&0xFF);
        c[2] = (char)((i>>8)&0xFF);
        c[3] = (char)(i&0xFF);
        return new String(c);
    }

    private static String bufferString(ByteBuffer buf, int offset) throws IOException {
        return new String(buf.array(), buf.arrayOffset() + offset, buf.limit()-offset, "UTF-8");
    }

    private void parse(Box box, String prefix, FieldMap fields) throws IOException {
        while (box != null) {
            Box databox = box;
            boolean descend = false;
//            descend = box.isContainer();
            String name = box.getName();
            if (name.equals("moov") || name.equals("udta") || name.equals("meta") || name.equals("ilst") || name.equals("trak") || name.equals("minf") || name.equals("stbl") || name.equals("mdia") || name.equals("stsd") || name.equals("mp4a")) {
                descend = true;
            } else if ("moov.udta.meta.ilst".equals(prefix)) {
                if (name.equals("----")) {
                    Box meanbox = databox.first();
                    Box namebox = meanbox.next();
                    name = bufferString(meanbox.getBuffer(), 4)+ "." + bufferString(namebox.getBuffer(), 4);
                    databox = namebox.next();
                } else {
                    databox = databox.first();
                }
            }
            if (prefix != null) {
                name = prefix + "." + name;
            }
//            System.out.println(Base64.toPrintableString(name)+" "+box+":"+databox+" "+Integer.toHexString(box.getType()));

            if (name.equals("moov.mvhd")) {
                ByteBuffer buf = databox.getBuffer();
                buf.position(0);
                int version = buf.getInt();
                version >>= 24;
                if (version == 0) {
                    int ctime = buf.getInt() + 2082870000;
                    int mtime = buf.getInt() + 2082870000;
                    int timescale = buf.getInt();
                    long duration = buf.getInt() * 1000L / timescale;
                    fields.put(Field.Duration, Integer.valueOf((int)duration));
                    fields.put(Field.Added, Integer.valueOf(ctime));
                    fields.put(Field.LastModified, Integer.valueOf(mtime));
                }
            } else if (name.equals("moov.trak.mdia.mdhd")) {
                ByteBuffer buf = databox.getBuffer();
                fields.put(Field.SampleRate, buf.getInt(1));
            } else if (name.equals("moov.trak.mdia.minf.stbl.stsd.mp4a.esds")) {
                ByteBuffer buf = databox.getBuffer();
                buf.position(0);
                int version = buf.getInt();
                while (buf.remaining() > 0) {
                    int section = buf.get();
//                    System.out.println("v="+version+" r="+buf.remaining()+" s="+section);
                    int len = buf.get() & 0xFF;
                    if ((len&0xFF) == 0x80 || (len&0xFF) == 0x81 || (len&0xFF) == 0xFE) {
                        buf.getShort();
                        len = buf.get();
                    }
                    if (section == 3) {
                        buf.position(buf.position() + 3);
                    } else if (section == 4) {
                        int ituneskind = buf.get();
                        buf.position(buf.position() + 4);
                        int maxbitrate = buf.getInt();
                        int avgbitrate = buf.getInt();
                        fields.put(Field.BitRate, Integer.valueOf(avgbitrate / 1000));
                    } else if (section == 5) {
                        int audioprofile = (buf.get() & 0xFF) >> 3;
                        if (audioprofile == 1) {
                            fields.put(Field.Codec, "Main-AAC");
                        } else if (audioprofile == 2) {
                            fields.put(Field.Codec, "LC-AAC");
                        } else if (audioprofile == 5) {
                            fields.put(Field.Codec, "HE-AAC");
                        } else if (audioprofile == 29) {
                            fields.put(Field.Codec, "HE-AACv2");
                        }
                        int channels = (buf.get() >> 3) & 0x7F;
                        if (channels == 1) {
                            fields.put(Field.Mono, Boolean.TRUE);
                        }
                    } else {
                        break;
                    }
                }
            } else if (name.equals("moov.trak.mdia.minf.stbl.stsd.alac")) {
                ByteBuffer buf = databox.getBuffer();
                buf.position(48);
                int channels = buf.getShort() & 0xFF;
                buf.getShort();
                buf.getInt();
                int bitrate = buf.getInt();
                int samplerate = buf.getInt();
                if (channels == 1) {
                    fields.put(Field.Mono, Boolean.TRUE);
                }
                fields.put(Field.Codec, "ALAC");
                fields.put(Field.BitRate, Integer.valueOf(bitrate / 1000));
                fields.put(Field.SampleRate, Integer.valueOf(samplerate));
            } else if (name.equals("mdat")) {
                long off = databox.first().offset() + 8;
                long len = databox.first().length() - 8;
                fields.put(Field.AudioByteRange, Long.valueOf(off<<32 + len));
            } else {
                Field field = tag2field.get(name);
                if (field != null) {
                    ByteBuffer buf = databox.getBuffer();
                    if (field == Field.Year) {
                        String t = bufferString(buf, 8);
                        if (t.length() >=4 && Character.isDigit(t.charAt(0)) && Character.isDigit(t.charAt(1)) && Character.isDigit(t.charAt(2)) && Character.isDigit(t.charAt(3))) {
                            fields.put(Field.Year, Integer.parseInt(t.substring(0, 4)));
                        }
                    } else if (field == Field.Track) {
                        buf.getShort();
                        int track = buf.getShort();
                        int numtracks = buf.getShort();
                        fields.putUntyped(field, Integer.valueOf(track));
                        if (numtracks != 0) {
                            fields.put(Field.NumTracks, Integer.valueOf(numtracks));
                        }
                    } else if (field == Field.Disc) {
                        buf.getShort();
                        int disc = buf.getShort();
                        int numdiscs = buf.getShort();
                        fields.putUntyped(field, Integer.valueOf(disc));
                        if (numdiscs != 0) {
                            fields.put(Field.NumDiscs, Integer.valueOf(numdiscs));
                        }
                    } else if (field == Field.BPM) {
                        fields.put(Field.BPM, Float.valueOf(buf.getShort()));
                    } else if (field == Field.ImageOffset) {
                        fields.put(Field.ImageOffset, Integer.valueOf((int)databox.first().offset() + 8));
                        fields.put(Field.ImageLength, Integer.valueOf(buf.limit()));
                    } else if (field == Field.Gain) {
                        String val = bufferString(buf, 8);
                        String[] com = val.trim().split(" ");
                        float lval1 = Integer.parseInt(com[0], 16) / 1000f;
                        float rval1 = Integer.parseInt(com[1], 16) / 1000f;
                        float lval2 = Integer.parseInt(com[2], 16) / 2500f;
                        float rval2 = Integer.parseInt(com[3], 16) / 2500f;
                        float gain = (float)(-10 * Math.log(Math.max(Math.max(lval1, lval2), Math.max(rval1, rval2))) / Math.log(10));
                        if (gain >= -25 && gain <= 25) {
                            fields.put(Field.Gain, Float.valueOf(gain));
                        }
                    } else if (field.getType() == String.class) {
                        String value;
                        if (field == Field.Genre && name.endsWith("gnre")) {
                            value = TrackReader.MP3GENRES[buf.getShort() - 1];
                        } else {
                            value = bufferString(buf, 8);
                        }
                        fields.putUntyped(field, value);
                    } else if (field.getType() == Boolean.class) {
                        if (buf.get() != 0) {
                            fields.putUntyped(field, Boolean.TRUE);
                        }
                    }
                }
            }

            if (descend) {
                parse(box.first(), name, fields);
            }
            box = box.next();
        }
    }

    private Box getBox(FileChannel channel, long off, long extent, Box parent) throws IOException {
        ByteBuffer lentype = ByteBuffer.allocate(8);
        channel.position(off);
        lentype.clear();
        channel.read(lentype);
        lentype.flip();
        long len = lentype.getInt();
        int type = lentype.getInt();
        if (len == 0) {
            len = channel.size() - off;
        } else if (len == 1) {
            lentype.clear();
            channel.read(lentype);
            len = lentype.getLong(0);
        }

        return new Box(type, channel, off, len, extent, parent);
    }

    private class Box {

        private FileChannel channel;
        private long off, len, extent;
        private int type;
        private Box next, first, parent;

        Box(int type, FileChannel channel, long off, long len, long extent, Box parent) {
            this.channel = channel;
            this.off = off;
            this.len = len;
            this.extent = extent;
            this.type = type;
            this.parent = parent;
        }

        int getFirstOffset() {
            String fullname = getFullName();
            int off = 8;
            if (fullname.equals("moov.udta.meta")) {
                off += 4;
            } else if (fullname.equals("moov.trak.mdia.minf.stbl.stsd")) {
                off += 8;
            } else if (fullname.equals("moov.trak.mdia.minf.stbl.stsd.mp4a")) {
                off += 28;
            }
            return off;
        }

        long offset() {
            return off;
        }

        long length() {
            return len;
        }

        int getType() {
            return type;
        }

        Box next() throws IOException {
            if (next == null) {
                if (len < extent) {
                    next = getBox(channel, off+len, extent-len, parent);
                } else {
                    next = NULL;
                }
            }
            return next==NULL ? null : next;
        }

        Box first() throws IOException {
            if (first == null) {
                int firstoff = getFirstOffset();
                if (len >= firstoff) {
                    first = getBox(channel, off+firstoff, len-firstoff, this);
                } else {
                    first = NULL;
                }
            }
            return first==NULL ? null : first;
        }

        String getFullName() {
           if (parent == null) {
               return typeString(type);
           } else {
               return parent.getFullName() + "." + typeString(type);
           }
        }

        String getName() {
           return typeString(type);
        }

        boolean isContainer() {
            for (int i=0;i<containers.length;i++) {
                if (type == containers[i]) {
                    return true;
                }
            }
            return false;
        }

        ByteBuffer getBuffer() throws IOException {
            ByteBuffer buf = ByteBuffer.allocate((int)(len - 8));
            channel.position(off + 8);
            channel.read(buf);
            buf.flip();
            buf.position(8);
            return buf;
        }

        public String toString() {
            return "["+getName()+" "+offset()+" "+length()+"]";
        }
    }

    public Map<String,Object> reportState() {
        Map<String,Object> map = new LinkedHashMap<String,Object>();
        return map;
    }

    public void startService(ServiceContext context) {
        this.context = context;
    }

    public void stopService(ServiceContext context) {
    }

    public ServiceContext getContext() {
        return context;
    }

    public static void main(String[] args) throws Exception {
        MP4Handler handler = new MP4Handler();
        for (int i=0;i<args.length;i++) {
            FieldMap map = handler.read(new File(args[i]));
            System.out.println(map);
            handler.getAudioInputStream(new File(args[i]));
        }
    }

    public AudioInputStream getAudioInputStream(File file) throws UnsupportedAudioFileException, IOException {
        AudioFileReader aac;
        try {
            aac = (AudioFileReader)Class.forName("net.sourceforge.jaad.spi.javasound.AACAudioFileReader").newInstance();
        } catch (Throwable e) {
            throw new UnsupportedAudioFileException("No AAC decoding, add \"jaadec\" package to classpath");
        }
        return aac.getAudioInputStream(file);
    }

}
