package org.qtunes.player;

import org.qtunes.core.*;
import org.qtunes.db.*;
import org.qtunes.auth.*;
import org.qtunes.speaker.*;
import java.util.*;
import java.net.URL;
import java.io.*;
import javax.sound.sampled.*;

/**
 * A Player plays tracks to one or more speakers which may be local speaker
 * (probably a {@link SourceDataLine}), an Airtunes speaker or some other type (eg RTSP, uPnP etc.)
 * Each player maintains it's own playlist and tracks may be cued, reordered, skipped
 * and so on as you'd expect.
 */
public interface Player extends Service {

    /**
     * Return the Playlist in use by this Player
     */
    @ControlMethod (
        params = {},
        auth = "play",
        help = "Return the playlist in use by this player"
    )
    public ManualPlaylist getPlaylist();

    /**
     * Skip to the next track in the playlist. If there are no
     * more tracks in the queue, stop playing
     */
    @ControlMethod (
        params = {},
        auth = "play",
        help = "Skip to the next Track"
    )
    public void playNext();

    /**
     * Skip to the previous track in the playlist. If you are currently
     * playing the first track, restart it.
     */
    @ControlMethod (
        params = {},
        auth = "play",
        help = "Skip to the previous Track"
    )
    public void playPrevious();

    /**
     * Set the currently playing track
     */
     @ControlMethod (
         params = {"index"},
         auth = "play",
         help = "Play the Track at the specifed position in the playlist"
     )
     public void setCurrentTrackIndex(int ix);

    /**
     * Seek to the specified millisecond value in the current track
     * If the value is outsize the valid range this method does nothing
     */
    @ControlMethod (
        params = {"ms"},
        auth = "play",
        help = "Seek to the specified milliseconds in the current Track"
    )
    public void seek(int ms);

    /**
     * Set the player to be paused (true) or playing (false).
     */
    @ControlMethod (
        params = {"paused"},
        auth = "play",
        help = "Pause or unpause the Player"
    )
    public void setPaused(boolean paused);

    /**
     * Return true if the player is paused, false if it's playing or
     * stopped.
     */
    @ControlMethod (
        params = {},
        help = "Return whether the Player is paused"
    )
    public boolean isPaused();

    /**
     * Return the number of milliseconds into the currently playing
     * track.
     */
    @ControlMethod (
        params = {},
        help = "Return the number of milliseconds into the currently playing Track"
    )
    public int getTrackPosition();

    /**
     * Return the number of milliseconds remaining in the playlist.
     */
    @ControlMethod (
        params = {},
        help = "Return the number of milliseconds remaining in the Playlist"
    )
    public long getRemaining();

    /**
     * Set whether to shuffle the playlist
     */
    @ControlMethod (
        params = {"shuffle"},
        auth = "play",
        help = "Set whether the Player is to shuffle or not"
    )
    public void setShuffle(boolean shuffle);

    /**
     * Set the style of repeat - none (0), track only (1) or playlist (2)
     */
    @ControlMethod (
        params = {"repeat"},
        auth = "play",
        help = "Set the Player repeat mode: none(0), track only(1) or playlist(2)"
    )
    public void setRepeat(int repeat);

    /**
     * Get the repeat mode, as set by {@link #setRepeat}
     */
    @ControlMethod (
        params = {},
        help = "Return the Repeat mode - none(0), track only(1) or playlist (2)"
    )
    public int getRepeat();

    /**
     * Get the shuffle mode, as set by {@link #setShuffle}
     */
    @ControlMethod (
        params = {},
        help = "Return whether the Player is set to shuffle"
    )
    public boolean getShuffle();

    /**
     * Return the index in the list of tracks returned from {@link #getPlaylist}
     * that is currently playing. If no track is playing return -1
     */
    @ControlMethod (
        params = {},
        help = "Return the position in the playlist of the currently playing Track, or -1 if no Track is playing"
    )
    public int getCurrentTrackIndex();

    /**
     * Set the current volume
     * @param volume the volume, between 0 and 100
     */
    @ControlMethod (
        params = {"volume"},
        auth = "play",
        help = "Set the volume, between 0 and 100"
    )
    public void setVolume(int volume);

    /**
     * Return the current volume
     * @return the volume, between 0 and 100
     */
    @ControlMethod (
        params = {},
        help = "Return the volume, between 0 and 100"
    )
    public int getVolume();

    //-------------------------------------------------------------------------------

    /**
     * Get the collection of speakers that could be used by this Player and that
     * aren't currently in use elsewhere.
     */
    @ControlMethod (
        params = {},
        auth = "play",
        help = "List the names of the Speakers available to this Player"
    )
    public Collection<Speaker> getAvailableSpeakers();

    /**
     * Add a Speaker to the Player. Should only be called by {@link Speaker#setPlayer}.
     */
    public void addSpeaker(Speaker speaker);

    /**
     * Remove a Speaker to the Player. Should only be called by {@link Speaker#setPlayer}.
     */
    public void removeSpeaker(Speaker speaker);

    /**
     * Return the list of {@link Speaker} objects current in use by this Player
     */
    @ControlMethod (
        params = {},
        auth = "play",
        help = "Get the list of speakers in use by this Player"
    )
    public Collection<Speaker> getSpeakers();

    /**
     * Get the {@link Track} currently being played by this Player, or null if it's
     * not playing anything
     */
    public Track getCurrentTrack();

    /**
     * Return the revision number. This should be increased every time the Player's state is
     * updated.
     */
    public int getRevision();

    /**
     * Return a nice name for this Player for display to the user
     */
    public String getDisplayName();

}
