package org.qtunes.scanner;

import org.qtunes.core.*;
import org.qtunes.db.*;
import org.qtunes.ff.*;
import java.io.*;
import java.util.*;
import java.util.regex.*;
import java.util.concurrent.*;

/**
 * The Scanner scans the disk looking for new files, and attempts
 * to load them by calling {@link TrackReader#read}. If it
 * succeeds it adds the tracks to the database.
 */
public class Scanner implements Service, Runnable {

    private ServiceContext context;
    private File root;
    private String rootpath;
    private volatile boolean cancelled;
    private boolean fullscan;
    private int rescan;
    private long lastscan;
    private Thread thread;
    private Pattern skip;

    public void startService(ServiceContext context) {
        this.context = context;
        rescan = context.getProperty("rescan")==null ? 0 : Integer.parseInt(context.getProperty("rescan"));
        fullscan = !"quick".equals(context.getProperty("scantype"));
        if (context.getProperty("lastscan")!=null) {
            lastscan = Long.parseLong(context.getProperty("lastscan"));
        }
        if (context.getProperty("skip")!=null) {
	    try {
		skip = Pattern.compile(context.getProperty("skip"));
	    } catch (Exception e) {
	        context.warn("Can't parse \"skip\" of \""+context.getProperty("skip")+"\"", e);
	    }
        }
        try {
            root = (new File(context.getGlobalProperty("root"))).getCanonicalFile();
            rootpath = root.getCanonicalPath();
            if (!root.canRead() || !root.isDirectory()) {
                throw new IOException("Cannot read root directory \""+root+"\"");
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
        thread = new Thread(this, "qTunes-"+context.getServiceName());
        thread.setPriority(Thread.MIN_PRIORITY);
        thread.start();
    }

    public void stopService(ServiceContext context) {
        cancelled = true;
        thread.interrupt();
        try {
            thread.join();
        } catch (InterruptedException e) { }
    }

    public ServiceContext getContext() {
        return context;
    }

    /**
     * Rescan for any tracks changed since the last scan
     */
    @ControlMethod (
        params = {},
        help = "Rescan for Tracks changed since the last scan"
    )
    public synchronized void rescan() {
        thread.interrupt();
    }

    /**
     * Rescan for any tracks changed since the specified timestamp
     * @param lastscan the timestamp in milliseconds
     */
    @ControlMethod (
        params = {"timestamp"},
        help = "Rescan for Tracks changed since the specified timestamp"
    )
    public synchronized void rescan(long lastscan) {
        this.lastscan = lastscan;
        thread.interrupt();
    }

    /**
     * Return a list of paths this Scanner will traverse
     */
    public File getRoot() {
        return root;
    }

    /**
     * Return the timestamp of the last scan in milliseconds
     */
    @ControlMethod (
        params = {},
        help = "Return the timestamp of the last scan"
    )
    public long getLastScan() {
        return this.lastscan;
    }

    /**
     * The main scanning thread
     */
    public void run() {
        boolean notfirst = false;
        Database db = context.getService(Database.class);

        while (!cancelled) {
            try {
                Collection<Track> all = db.getTracks();
                long newlast = System.currentTimeMillis();
                Set changed = new HashSet();
                Set<String> filenames = null;
                Collection<Track> removed = null;
                context.info("Scan started: lastscan="+(new Date(lastscan))+" full="+fullscan+" root="+root);

                if (fullscan) {
                    filenames = new HashSet<String>(db.size());
                    removed = new HashSet<Track>();
                    for (Iterator<Track> i = all.iterator();!cancelled && i.hasNext();) {
                        Track track = i.next();
                        File file = track.getFile();
                        if (!file.exists()) {
                            removed.add(track);
                        } else {
                            filenames.add(track.get(Field.File));
                        }
                    }
                }

                findChanged(root, removed, filenames);
                filenames = null;
                if (cancelled) {
                    context.info("Scan cancelled");
                } else {
                    context.info("Scan completed");
                    if (removed != null) {
                        for (Iterator<Track> i = removed.iterator();i.hasNext();) {
                            i.next().delete();
                        }
                    }
                    db.commit();
                    context.fireEvent("completed", null);
                    lastscan = newlast;
                    context.putProperty("lastscan", Long.toString(lastscan));
                }
            } catch (IOException e) {
                context.warn("Scan failed", e);
            }
            try {
                Thread.sleep(rescan==0 ? Integer.MAX_VALUE : rescan*1000);
            } catch (InterruptedException e) { }
        }
    }

    private void addChanged(FieldMap fields, Collection<Track> removed) {
        Database db = context.getService(Database.class);
        Track track = db.create(fields);
        if (removed != null) {
            removed.remove(track);
        }
    }

    private void findChanged(File file, Collection<Track> removed, Set<String> filenames) throws IOException {
        if (!cancelled && file.canRead()) {
            if (file.isDirectory()) {
                File[] files = file.listFiles();
                for (int i=0;i<files.length;i++) {
                    findChanged(files[i], removed, filenames);
                }
            } else {
                String filename = file.getCanonicalPath();
                if (filename.startsWith(rootpath)) {
                    filename = "$" + filename.substring(rootpath.length());
                }
		if ((skip == null || !skip.matcher(filename).find()) && ((filenames != null && !filenames.contains(filename)) || file.lastModified() > lastscan)) {
                    try {
                        TrackReader trackreader = TrackReader.getTrackReader(context, file);
                        if (trackreader != null) {
                            FieldMap fields = trackreader.read(file);
                            fields.put(Field.File, filename);
                            addChanged(fields, removed);
                        }
                    } catch (IOException e) {
                        context.warn("Exception reading "+file, e);
                    }
                }
            }
        }
    }

    public Map<String,Object> reportState() {
        return null;
//        Map<String,Object> map = new LinkedHashMap<String,Object>();
//        map.put("lastscan", Long.parseLong(context.getProperty("lastscan")));
//        map.put("roots", getPaths());
//        return map;
    }

}
