package org.qtunes.web;

import java.io.*;
import java.util.*;
import java.text.*;
import java.util.zip.*;
import java.net.*;

/**
 * A wrapping WebHandler that will optionally cache and compress
 * the response from it's handler.
 */
public class CachingWebHandler implements WebHandler {

    private final WebHandler handler;
    private final Set<URI> seen;

    /**
     * Create a Caching WebHandler that will wrap the specified handler
     */
    public CachingWebHandler(WebHandler handler) {
        this.handler = handler;
        this.seen = new HashSet<URI>();
    }

    public void handle(WebConnection con) throws IOException {
        DateFormat dateformat = new SimpleDateFormat(DATEFORMAT);

        String ifs = con.getRequestHeader("If-Modified-Since");
        URI uri = con.getRequestURI();
        if (ifs != null && seen.contains(uri)) {
            try {
                Date date = dateformat.parse(ifs);
                if (!isModifiedSince(uri, date)) {
                    con.sendResponseHeaders(304, -1);
                    return;
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        con.setResponseHeader("Date", new Date());
        if (isCompressable(con)) {
            handler.handle(new CompressingWebConnection(con));
        } else {
            handler.handle(con);
        }
        seen.add(uri);
    }

    /**
     * Return true if the request has been modified since the specified Date
     * and a 304 can be returned. By default this method returns "true" - subclasses
     * should override this method.
     *
     * @param uri the requested URI
     * @param date the Date of the last copy of this URI on the client
     * @return true if the resource for this URI has been modified
     */
    protected boolean isModifiedSince(URI uri, Date date) {
        return true;
    }

    /**
     * Return true if the response to the specified WebConnection should
     * compress content of type "application/json" or "text/*". By default
     * this method returns true. Other MIME-types are never compressed.
     */
    protected boolean isCompressable(WebConnection con) {
        return true;
    }

    private class CompressingWebConnection extends WebConnectionWrapper {

        private ByteArrayOutputStream out;
        private GZIPOutputStream zout;
        private int status;

        CompressingWebConnection(WebConnection con) {
            super(con);
            out = new ByteArrayOutputStream();
        }

        public void close() throws IOException {
            if (zout != null) {
                GZIPOutputStream t = zout;
                zout = null;
                t.close();
                super.setResponseHeader("Content-Encoding", "gzip");
                super.sendResponseHeaders(status, out.size());
                OutputStream sout = super.getOutputStream();
                out.writeTo(sout);
                sout.close();
            }
        }

        public void sendResponseHeaders(int status, int length) throws IOException {
            if (out == null || length == -1) {
                super.sendResponseHeaders(status, length);
                out = null;
            } else {
                this.status = status;
            }
        }

        public OutputStream getOutputStream() {
            if (out == null) {
                return super.getOutputStream();
            } else if (zout == null) {
                try {
                    zout = new GZIPOutputStream(out) {
                        public void close() throws IOException {
                            super.close();
                            CompressingWebConnection.this.close();
                        }
                    };
                } catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }
            return zout;
        }

        private void testResponseHeader(String key, String value) {
            if (key.equalsIgnoreCase("Content-Encoding") && value.equals("gzip")) {
                zout = null;    // Already compressed
                out = null;
            } else if (key.equalsIgnoreCase("Content-Type")) {
                int ix = value.indexOf(";");
                if (ix > 0) {
                    value = value.substring(0, ix);
                }
                if (value != null && !value.equals("application/json") && !value.startsWith("text/")) {
                    zout = null;    // Not text, don't compress
                    out = null;
                }
            }
        }

        public void addResponseHeader(String key, String value) {
            testResponseHeader(key, value);
            super.setResponseHeader(key, value);
        }

        public void setResponseHeader(String key, String value) {
            testResponseHeader(key, value);
            super.setResponseHeader(key, value);
        }

    }

}
