package org.qtunes.web;

import java.io.*;
import java.net.*;
import java.util.*;
import org.qtunes.web.*;
import org.qtunes.core.*;

/**
 * A StaticWebHandler returns a static resource. Caching is done
 * internally - it's assumed resources are loaded from Jars so are
 * unchanging. Pre-compressed versions of a resource may be stored
 * with a ".gz" suffix and will be returned if found.
 */
public class StaticWebHandler implements WebHandler {

    private Date lastmodified;
    private final String oldprefix, newprefix;
    private final Class base;
    private final ServiceContext context;
    private Set<String> seen;
    private String encoding = ";charset=UTF-8";
    private ThreadLocal<byte[]> cache = new ThreadLocal<byte[]>() {
        protected byte[] initialValue() {
            return new byte[8192];
        }
    };

    /**
     * Create a new StaticWebHandler
     * @param context the ServiceContext
     * @param oldprefix the prefix to the incoming URI that should be matched and removed
     * @param newprefix the prefix to the resource that should replace oldprefix
     * @param base the base class from which resources should be loaded
     */
    public StaticWebHandler(ServiceContext context, String oldprefix, String newprefix, Class base) {
        this.context = context;
        this.oldprefix = oldprefix;
        this.newprefix = newprefix;
        this.base = base;
        this.seen = new HashSet<String>();
        this.lastmodified = context.getServerBuildDate();
        if (lastmodified == null) {
            lastmodified = new Date();
        }
    }

    /**
     * Set the encoding for any text resources (txt, html, css, js) that are
     * served by this handler. The default encoding is "UTF-8"
     */
    public void setEncoding(String encoding) {
        if ("ISO-8859-1".equals(encoding)) {
            this.encoding = "";
        } else {
            this.encoding = ";charset="+encoding;
        }
    }

    public String getPrefix() {
        return oldprefix;
    }

    public void handle(WebConnection con) throws IOException {
        URI uri = con.getRequestURI();
        String path = uri.getPath();
        if (path.startsWith(oldprefix)) {
            path = newprefix + path.substring(oldprefix.length());

            URL url = base.getResource(path+".gz");
            if (url != null) {
                con.setResponseHeader("Content-Encoding", "gzip");
            } else {
                url = base.getResource(path);
            }
            if (url == null) {
                con.sendResponseHeaders(404, -1);
            } else {
                String ifs = con.getRequestHeader("If-Modified-Since");
                if (ifs != null && seen.contains(path)) {
                    con.sendResponseHeaders(304, -1);
                } else {
                    URLConnection urlcon = url.openConnection();
                    String type;
                    if (path.endsWith(".css")) {
                        type = "text/css"+encoding;
                    } else if (path.endsWith(".js")) {
                        type = "text/javascript"+encoding;
                    } else if (path.endsWith(".png")) {
                        type = "image/png";
                    } else if (path.endsWith(".jpg")) {
                        type = "image/jpg";
                    } else if (path.endsWith(".gif")) {
                        type = "image/gif";
                    } else if (path.endsWith(".swf")) {
                        type = "application/x-shockwave-flash";
                    } else if (path.endsWith(".jar")) {
                        type = "application/x-java-applet";
                    } else if (path.endsWith(".txt")) {
                        type = "text/plain"+encoding;
                    } else {
                        type = "text/html"+encoding;
                    }
                    con.setResponseHeader("Content-Type", type);
                    con.setResponseHeader("Date", new Date());
                    con.setResponseHeader("Last-Modified", lastmodified);
                    if (urlcon.getContentLength() > 0) {
                        con.sendResponseHeaders(200, urlcon.getContentLength());
                    } else {
                        con.sendResponseHeaders(200, 0);
                    }
                    BufferedInputStream in = new BufferedInputStream(urlcon.getInputStream());
                    OutputStream out = con.getOutputStream();
                    byte[] buf = cache.get();
                    int len;
                    try {
                        while ((len=in.read(buf))>=0) {
                            out.write(buf, 0, len);
                        }
                    } finally {
                        in.close();
                        out.close();
                    }
                }
            }
        } else {
            throw new IllegalArgumentException("Wrong prefix");
        }
    }

}
